import io
import textwrap
import unittest

from okonomiyaki.versions import EnpkgVersion

from simplesat.constraints import (
    ConstraintModifiers, PrettyPackageStringParser, InstallRequirement
)
from simplesat.dependency_solver import (
    DependencySolver, packages_are_consistent,
    requirements_from_packages, packages_from_requirements,
    requirements_are_satisfiable, requirements_are_complete,
    satisfy_requirements, simplify_requirements,
)
from simplesat.errors import (
    MissingInstallRequires, SatisfiabilityError, SatisfiabilityErrorWithHint
)
from simplesat.pool import Pool
from simplesat.repository import Repository
from simplesat.request import Request
from simplesat.test_utils import Scenario
from simplesat.transaction import (
    InstallOperation, RemoveOperation, UpdateOperation
)


R = InstallRequirement._from_string
P = PrettyPackageStringParser(EnpkgVersion.from_string).parse_to_package


class SolverHelpersMixin(object):
    def setUp(self):
        self.repository = Repository()
        self.installed_repository = Repository()

        self._package_parser = PrettyPackageStringParser(
            EnpkgVersion.from_string
        )

    def package_factory(self, s):
        return self._package_parser.parse_to_package(s)

    def resolve(self, request, strict=False):
        pool = Pool([self.repository, self.installed_repository])
        solver = DependencySolver(
            pool, [self.repository], self.installed_repository,
            use_pruning=False, strict=strict
        )
        return solver.solve(request)

    def resolve_with_hint(self, request, strict=False):
        pool = Pool([self.repository, self.installed_repository])
        solver = DependencySolver(
            pool, [self.repository], self.installed_repository,
            use_pruning=False, strict=strict
        )
        return solver.solve_with_hint(request)

    def assertEqualOperations(self, operations, r_operations):
        self.assertEqual(operations, r_operations)


class TestSolver(SolverHelpersMixin, unittest.TestCase):
    def test_simple_install(self):
        # Given
        mkl = self.package_factory(u"mkl 10.3-1")
        self.repository.add_package(mkl)

        r_operations = [InstallOperation(mkl)]

        request = Request()
        request.install(R("mkl"))

        # When
        transaction = self.resolve(request)

        # Then
        self.assertEqualOperations(transaction.operations, r_operations)

    def test_multiple_installs(self):
        # Given
        mkl = self.package_factory(u"mkl 10.3-1")
        libgfortran = self.package_factory(u"libgfortran 3.0.0-2")

        r_operations = [
            InstallOperation(libgfortran),
            InstallOperation(mkl),
        ]

        self.repository.add_package(mkl)
        self.repository.add_package(libgfortran)

        request = Request()
        request.install(R("mkl"))
        request.install(R("libgfortran"))

        # When
        transaction = self.resolve(request)

        # Then
        self.assertEqualOperations(transaction.operations, r_operations)

    def test_simple_dependency(self):
        # Given
        mkl = self.package_factory(u"mkl 10.3-1")
        libgfortran = self.package_factory(u"libgfortran 3.0.0-2")
        numpy = self.package_factory(
            u"numpy 1.9.2-1; depends (mkl == 10.3-1, libgfortran ^= 3.0.0)"
        )

        r_operations = [
            # libgfortran sorts before mkl
            InstallOperation(libgfortran),
            InstallOperation(mkl),
            InstallOperation(numpy),
        ]

        self.repository.add_package(mkl)
        self.repository.add_package(libgfortran)
        self.repository.add_package(numpy)

        request = Request()
        request.install(R("numpy"))

        # When
        transaction = self.resolve(request)

        # Then
        self.assertEqualOperations(transaction.operations, r_operations)

    def test_already_installed(self):
        # Given
        mkl1 = self.package_factory(u"mkl 10.3-1")
        mkl2 = self.package_factory(u"mkl 10.3-2")

        r_operations = []

        self.repository.add_package(mkl1)
        self.repository.add_package(mkl2)
        self.installed_repository.add_package(mkl1)

        # When
        request = Request()
        request.install(R("mkl"))

        transaction = self.resolve(request)

        # Then
        self.assertEqualOperations(transaction.operations, r_operations)

        # Given
        r_operations = [
            RemoveOperation(mkl1),
            InstallOperation(mkl2),
        ]
        r_pretty_operations = [
            UpdateOperation(mkl2, mkl1),
        ]

        # When
        request = Request()
        request.install(R("mkl > 10.3-1"))

        # When
        transaction = self.resolve(request)

        # Then
        self.assertEqualOperations(transaction.operations, r_operations)
        self.assertEqualOperations(
            transaction.pretty_operations, r_pretty_operations)

    def test_requirements_are_satisfiable(self):
        # Given
        scenario = Scenario.from_yaml(io.StringIO(u"""
                packages:
                    - MKL 10.2-1
                    - MKL 10.3-1
                    - numpy 1.7.1-1; depends (MKL == 10.3-1)
                    - numpy 1.8.1-1; depends (MKL == 10.3-1)

                request:
                    - operation: "install"
                      requirement: "numpy"
        """))
        packages = tuple(p for r in scenario.remote_repositories for p in r)
        requirements = [job.requirement for job in scenario.request.jobs]

        # When
        result = requirements_are_satisfiable(packages, requirements)

        # Then
        self.assertTrue(result.is_satisfiable)
        self.assertEqual(result.message, "")

    def test_requirements_are_not_satisfiable(self):
        # Given
        scenario = Scenario.from_yaml(io.StringIO(u"""
            packages:
                - MKL 10.2-1
                - MKL 10.3-1
                - numpy 1.7.1-1; depends (MKL == 10.3-1)
                - numpy 1.8.1-1; depends (MKL == 10.3-1)

            request:
                - operation: "install"
                  requirement: "numpy"
                - operation: "install"
                  requirement: "MKL != 10.3-1"
        """))
        r_msg = textwrap.dedent("""\
        Conflicting requirements:
        Requirements: 'numpy' <- 'MKL == 10.3-1' <- 'MKL'
            Can only install one of: (+MKL-10.3-1 | +MKL-10.2-1)
        Requirements: 'numpy' <- 'MKL == 10.3-1'
            numpy-1.8.1-1 requires (+MKL-10.3-1)
        Requirements: 'numpy'
            Install command rule (+numpy-1.7.1-1 | +numpy-1.8.1-1)
        """)
        packages = tuple(p for r in scenario.remote_repositories for p in r)
        requirements = [job.requirement for job in scenario.request.jobs]

        # When
        result = requirements_are_satisfiable(packages, requirements)

        # Then
        self.assertFalse(result.is_satisfiable)
        self.assertMultiLineEqual(result.message, r_msg)

    def test_requirements_are_complete(self):
        # Given
        scenario = Scenario.from_yaml(io.StringIO(u"""
            packages:
                - MKL 10.3-1
                - numpy 1.8.1-1; depends (MKL == 10.3-1)

            request:
                - operation: install
                  requirement: numpy ^= 1.8.1
                - operation: install
                  requirement: MKL == 10.3-1
        """))
        packages = tuple(p for r in scenario.remote_repositories for p in r)

        # When
        requirements = [job.requirement for job in scenario.request.jobs]
        result = requirements_are_complete(packages, requirements)

        # Then
        self.assertTrue(result.is_satisfiable)
        self.assertEqual(result.message, "")

    def test_requirements_are_not_complete(self):
        # Given
        scenario = Scenario.from_yaml(io.StringIO(u"""
            packages:
                - MKL 10.3-1
                - numpy 1.8.1-1; depends (MKL == 10.3-1)

            request:
                - operation: install
                  requirement: numpy
        """))
        r_msg = textwrap.dedent("""\
        Conflicting requirements:
        Requirements: 'numpy' <- 'MKL == 10.3-1'
            +numpy-1.8.1-1 was ignored because it depends on missing packages
        Requirements: 'numpy'
            Install command rule (+numpy-1.8.1-1)
        """)
        packages = tuple(p for r in scenario.remote_repositories for p in r)

        # When
        requirements = [job.requirement for job in scenario.request.jobs]
        result = requirements_are_complete(packages, requirements)

        # Then
        self.assertFalse(result.is_satisfiable)
        self.assertMultiLineEqual(result.message, r_msg)

    def test_packages_from_requirements(self):
        # Given
        requirements = (
            R(u'MKL == 10.3-1'),
            R(u'numpy == 1.9.1-1'),
            R(u'numpy == 1.9.1-2'),
            R(u'numpy == 1.9.1-3'),
        )
        expected = (
            P(u'MKL 10.3-1'),
            P(u'numpy 1.9.1-1; depends (MKL == 10.3-1)'),
            P(u'numpy 1.9.1-2; depends (MKL)'),
            P(u'numpy 1.9.1-3'),
        )

        # When
        repository = packages_from_requirements(expected, requirements)

        # Then
        result = tuple(repository)
        self.assertEqual(result, expected)

    def test_requirements_from_packages(self):
        # Given
        packages = (
            P(u'MKL 10.3-1'),
            P(u'numpy 1.9.1-1; depends (MKL == 10.3-1)'),
            P(u'numpy 1.9.1-2; depends (MKL)'),
            P(u'numpy 1.9.1-3'),
        )
        expected = (
            R(u'MKL == 10.3-1'),
            R(u'numpy == 1.9.1-1'),
            R(u'numpy == 1.9.1-2'),
            R(u'numpy == 1.9.1-3'),
        )

        # When
        result = requirements_from_packages(packages)

        # Then
        self.assertEqual(result, expected)

    def test_packages_are_consistent(self):
        # Given
        scenario = Scenario.from_yaml(io.StringIO(u"""
            packages:
                - MKL 10.3-1
                - numpy 1.8.1-1; depends (MKL == 10.3-1)
        """))

        # When
        packages = tuple(p for r in scenario.remote_repositories for p in r)
        result = packages_are_consistent(packages)

        # Then
        self.assertTrue(result)

    def test_repository_is_not_consistent(self):
        # Given
        scenario = Scenario.from_yaml(io.StringIO(u"""
            packages:
                - numpy 1.8.1-1; depends (MKL == 10.3-1)
        """))
        r_msg = textwrap.dedent("""\
        Conflicting requirements:
        Requirements: 'numpy == 1.8.1-1' <- 'MKL == 10.3-1'
            +numpy-1.8.1-1 was ignored because it depends on missing packages
        Requirements: 'numpy == 1.8.1-1'
            Install command rule (+numpy-1.8.1-1)
        """)

        # When
        packages = tuple(p for r in scenario.remote_repositories for p in r)
        result = packages_are_consistent(packages)

        # Then
        self.assertFalse(result.is_satisfiable)
        self.assertMultiLineEqual(result.message, r_msg)

    def test_missing_direct_dependency_fails(self):
        # Given
        numpy192 = self.package_factory(u"numpy 1.9.2-1")
        numpy200 = self.package_factory(u"numpy 2.0.0-1; depends (missing)")

        self.repository.add_package(numpy192)
        self.repository.add_package(numpy200)

        # When
        request = Request()
        request.install(R("numpy >= 2.0"))

        # Then
        with self.assertRaises(SatisfiabilityError):
            self.resolve(request)

    def test_missing_indirect_dependency_fails(self):
        # Given
        mkl = self.package_factory(u"MKL 10.3-1; depends (MISSING)")
        numpy192 = self.package_factory(u"numpy 1.9.2-1")
        numpy200 = self.package_factory(u"numpy 2.0.0-1; depends (MKL)")

        self.repository.add_package(mkl)
        self.repository.add_package(numpy192)
        self.repository.add_package(numpy200)

        # When
        request = Request()
        request.install(R("numpy >= 2.0"))

        # Then
        with self.assertRaises(SatisfiabilityError):
            self.resolve(request)

    def test_satisfy_requirements(self):
        requirements = (
            R(u'B ^= 1.0.0'),
            R(u'E'),
        )

        packages = (
            P(u"A 0.0.1-1;"),
            P(u"A 1.0.0-1;"),
            P(u"A 2.0.0-1;"),
            P(u"A 2.1.0-1;"),
            P(u"B 1.0.0-1; depends (A ^= 1.0, A != 2.1.0-1)"),
            P(u"B 2.0.0-1; depends (A != 2.1.0-1)"),
            P(u"C 1.0.0-1; depends (B > 2.0.0-1)"),
            P(u"D 1.0.0-1;"),
            P(u"E 1.0.0-1; depends (C > 2.0, D < 1.0)"),
        )

        modifiers = ConstraintModifiers(
            allow_newer=[u'A'],
            allow_older=[u'B'],
            allow_any=[u'C', u'D'],
        )

        # When
        result = satisfy_requirements(
            packages, requirements, modifiers=modifiers)
        expected = tuple(packages[i] for i in (2, 7, 4, 6, 8))

        # Then
        self.assertEqual(result, expected)

    def test_satisfy_requirements_fail(self):
        requirements = (
            R(u'B ^= 1.0.0'),
            R(u'E'),
        )

        packages = (
            P(u"A 0.0.1-1;"),
            P(u"A 1.0.0-1;"),
            P(u"A 2.0.0-1;"),
            P(u"A 2.1.0-1;"),
            P(u"B 1.0.0-1; depends (A ^= 1.0, A != 2.1.0-1)"),
            P(u"B 2.0.0-1; depends (A != 2.1.0-1)"),
            P(u"C 1.0.0-1; depends (B > 2.0.0-1)"),
            P(u"D 1.0.0-1;"),
            P(u"E 1.0.0-1; depends (C > 2.0, D < 1.0)"),
            P(u"F 1.0.0-1; depends (NONEXISTENT)"),
        )

        modifiers = ConstraintModifiers(
            allow_newer=[u'A'],
            allow_any=[u'C', u'D'],
        )

        # When / Then
        with self.assertRaises(SatisfiabilityError):
            satisfy_requirements(packages, requirements, modifiers=modifiers)

        with self.assertRaises(SatisfiabilityError):
            requirements = (R(u'F'),)
            satisfy_requirements(packages, requirements, modifiers=modifiers)

    def test_simplify_requirements(self):

        # Given
        requirements = (
            R(u'MKL == 10.3-1'),
            R(u'mismatch == 1.2.3-5'),
            R(u'numpy == 1.9.1-1'),
        )
        packages = (
            P(u'MKL 10.3-1'),
            P(u'numpy 1.9.1-1; depends (MKL == 10.3-1, mismatch == 1.2.3-4)'),
            P(u'mismatch 1.2.3-5'),
            P(u'unused 1.2.3-0'),
        )

        # When
        result = simplify_requirements(packages, requirements)
        expected = requirements[1:]

        # Then
        self.assertEqual(result, expected)

    def test_strange_key_error_bug_on_failure(self):
        # Given
        mkl = self.package_factory(u'MKL 10.3-1')
        libgfortran = self.package_factory(u'libgfortran 3.0.0-2')
        numpy192 = self.package_factory(
            u"numpy 1.9.2-1; depends (libgfortran ^= 3.0.0, MKL == 10.3-1)")
        numpy200 = self.package_factory(
            u"numpy 2.0.0-1; depends (nonexistent)")
        request = Request()

        # When
        for pkg in (mkl, libgfortran, numpy192, numpy200):
            self.repository.add_package(pkg)
        request.install(R("numpy >= 2.0"))

        # Then
        with self.assertRaises(SatisfiabilityError):
            self.resolve(request)

    def test_missing_dependency_strict(self):
        # Given
        mkl = self.package_factory(u'MKL 10.3-1')
        libgfortran = self.package_factory(u'libgfortran 3.0.0-2')
        numpy192 = self.package_factory(
            u"numpy 1.9.2-1; depends (libgfortran ^= 3.0.0, MKL == 10.3-1)")
        numpy200 = self.package_factory(
            u"numpy 2.0.0-1; depends (nonexistent)")
        request = Request()

        # When
        for pkg in (mkl, libgfortran, numpy192, numpy200):
            self.repository.add_package(pkg)
        request.install(R("numpy == 2.0.0-1"))

        # Then
        with self.assertRaises(MissingInstallRequires):
            self.resolve(request, strict=True)

    def test_upgrade_simple(self):
        # Given
        mkl_11_3_1 = P(u"mkl 11.3.1-1")
        mkl_2017_0_1_1 = P(u"mkl 2017.0.1-1")
        mkl_2017_0_1_2 = P(u"mkl 2017.0.1-2")

        numpy_1_10_4 = P(u"numpy 1.10.4-1; depends (mkl ^= 11.3.1)")
        numpy_1_11_3 = P(u"numpy 1.11.3-1; depends (mkl ^= 2017.0.1)")

        scipy_0_17_1 = P(u"scipy 0.17.1-1; depends (mkl ^= 11.3.1, numpy ^= 1.10.4)")
        scipy_0_18_1 = P(u"scipy 0.18.1-1; depends (mkl ^= 2017.0.1, numpy ^= 1.11.3)")

        self.repository.update([
            mkl_11_3_1, mkl_2017_0_1_1, mkl_2017_0_1_2, numpy_1_10_4,
            numpy_1_11_3, scipy_0_17_1, scipy_0_18_1
        ])
        self.installed_repository.update([mkl_11_3_1, numpy_1_10_4, scipy_0_17_1])

        r_operations = [
            RemoveOperation(scipy_0_17_1),
            RemoveOperation(numpy_1_10_4),
            RemoveOperation(mkl_11_3_1),
            InstallOperation(mkl_2017_0_1_2),
            InstallOperation(numpy_1_11_3),
            InstallOperation(scipy_0_18_1),
        ]

        # When
        request = Request()
        request.upgrade()

        transaction = self.resolve(request)

        # Then
        self.assertEqualOperations(transaction.operations, r_operations)

    def test_upgrade_fail(self):
        # Given
        mkl_11_3_1 = P(u"mkl 11.3.1-1")
        mkl_2017_0_1_1 = P(u"mkl 2017.0.1-1")
        mkl_2017_0_1_2 = P(u"mkl 2017.0.1-2")

        numpy_1_10_4 = P(u"numpy 1.10.4-1; depends (mkl ^= 11.3.1)")

        scipy_0_17_1 = P(u"scipy 0.17.1-1; depends (mkl ^= 11.3.1, numpy ^= 1.10.4)")

        self.repository.update([
            mkl_11_3_1, mkl_2017_0_1_1, mkl_2017_0_1_2, numpy_1_10_4,
            scipy_0_17_1
        ])
        self.installed_repository.update([mkl_11_3_1, numpy_1_10_4, scipy_0_17_1])

        # When/Then
        request = Request()
        request.upgrade()

        with self.assertRaises(SatisfiabilityError):
            self.resolve(request)

    def test_upgrade_no_candidate(self):
        # Given
        mkl_11_3_1 = P(u"mkl 11.3.1-1")
        mkl_2017_0_1_1 = P(u"mkl 2017.0.1-1")
        mkl_2017_0_1_2 = P(u"mkl 2017.0.1-2")

        numpy_1_10_4 = P(u"numpy 1.10.4-1; depends (mkl ^= 11.3.1)")

        scipy_0_17_1 = P(u"scipy 0.17.1-1; depends (mkl ^= 11.3.1, numpy ^= 1.10.4)")

        gnureadline_6_3 = P(u"gnureadline 6.3-1")

        # gnureadline is not available in the remote repository
        self.repository.update([
            mkl_11_3_1, mkl_2017_0_1_1, mkl_2017_0_1_2, numpy_1_10_4,
            scipy_0_17_1
        ])
        self.installed_repository.update([mkl_11_3_1, numpy_1_10_4, scipy_0_17_1])
        self.installed_repository.update([gnureadline_6_3])

        # When/Then
        request = Request()
        request.upgrade()

        with self.assertRaises(SatisfiabilityError):
            self.resolve(request)

    def test_up_to_date_package_choices(self):
        # Given
        packages = u"""
            envisage 4.7.2-1; depends (apptools ^= 4.4.0, traits ^= 5.1.1)
            apptools 4.4.0-12; depends (traitsui ^= 6.0.0)
            apptools 4.4.0-14; depends (traitsui ^= 6.1.1)
            traitsui 6.0.0-3; depends (traits ^= 5.1.0)
            traitsui 6.0.0-4; depends (traits ^= 5.1.1)
            traitsui 6.1.1-1; depends (traits ^= 5.1.1)
            traits 5.1.0-1
            traits 5.1.1-1
        """
        for package in packages.strip().splitlines():
            self.repository.add_package(P(package.strip()))

        # When
        request = Request()
        request.install(R("envisage"))
        transaction = self.resolve(request)
        packages = {
            operation.package.name: str(operation.package.version)
            for operation in transaction.operations
        }

        # Then
        expected_packages = {
            "traits": "5.1.1-1",
            "traitsui": "6.1.1-1",
            "apptools": "4.4.0-14",
            "envisage": "4.7.2-1",
        }
        self.assertEqual(packages, expected_packages)



class TestSolverWithHint(SolverHelpersMixin, unittest.TestCase):
    def test_no_conflict(self):
        # Given
        mkl_2017_0_1_1 = P(u"mkl 2017.0.1-1")
        numpy_1_11_3 = P(u"numpy 1.11.3-1; depends (mkl ^= 2017.0.1)")
        scipy_0_18_1 = P(u"scipy 0.18.1-1; depends (mkl ^= 2017.0.1, numpy ^= 1.11.3)")

        r_operations = [
            InstallOperation(mkl_2017_0_1_1),
            InstallOperation(numpy_1_11_3),
            InstallOperation(scipy_0_18_1),
        ]

        self.repository.update([mkl_2017_0_1_1, numpy_1_11_3, scipy_0_18_1])

        request = Request()
        request.install(R(u"scipy >= 0.18.0"))
        request.install(R(u"numpy >= 1.10.0"))


        # When
        transaction = self.resolve_with_hint(request)

        # Then
        self.assertEqualOperations(transaction.operations, r_operations)

    def test_direct_dependency_conflict(self):
        # Given
        mkl_11_3_1 = P(u"mkl 11.3.1-1")
        mkl_2017_0_1_1 = P(u"mkl 2017.0.1-1")
        numpy_1_11_3 = P(u"numpy 1.11.3-1; depends (mkl ^= 2017.0.1)")
        scipy_0_18_1 = P(u"scipy 0.18.1-1; depends (mkl ^= 2017.0.1, numpy ^= 1.11.3)")

        self.repository.update([
            mkl_11_3_1, mkl_2017_0_1_1, numpy_1_11_3, scipy_0_18_1,
        ])

        request = Request()
        request.install(R(u"scipy >= 0.18.0"))
        request.install(R(u"numpy >= 1.10.0"))
        request.install(R(u"mkl < 12"))  # MKL < 12 conflicts with scipy >= 0.18.0

        r_hint_pretty_string = textwrap.dedent(u"""\
            The following jobs are conflicting:
                install numpy >= 1.10.0-0
                install mkl < 12-0"""
        )


        # When/Then
        with self.assertRaises(SatisfiabilityErrorWithHint) as ctx:
            self.resolve_with_hint(request)

        self.assertMultiLineEqual(
            ctx.exception.hint_pretty_string, r_hint_pretty_string)

    def test_upgrade_fail(self):
        # Given
        mkl_11_3_1 = P(u"mkl 11.3.1-1")
        mkl_2017_0_1_1 = P(u"mkl 2017.0.1-1")
        mkl_2017_0_1_2 = P(u"mkl 2017.0.1-2")

        numpy_1_10_4 = P(u"numpy 1.10.4-1; depends (mkl ^= 11.3.1)")

        scipy_0_17_1 = P(u"scipy 0.17.1-1; depends (mkl ^= 11.3.1, numpy ^= 1.10.4)")

        self.repository.update([
            mkl_11_3_1, mkl_2017_0_1_1, mkl_2017_0_1_2, numpy_1_10_4,
            scipy_0_17_1
        ])
        self.installed_repository.update([mkl_11_3_1, numpy_1_10_4, scipy_0_17_1])

        r_hint_pretty_string = textwrap.dedent(u"""\
            The following jobs are conflicting:
                install mkl == 2017.0.1-2
                install numpy == 1.10.4-1"""
        )

        # When/Then
        request = Request()
        request.upgrade()

        with self.assertRaises(SatisfiabilityErrorWithHint) as ctx:
            self.resolve_with_hint(request)

        self.assertMultiLineEqual(
            ctx.exception.hint_pretty_string, r_hint_pretty_string)



import os.path
from unittest import TestCase

import six

from simplesat.errors import NoPackageFound, SatisfiabilityError
from simplesat.dependency_solver import DependencySolver
from simplesat.pool import Pool
from simplesat.test_utils import Scenario
from simplesat.transaction import (
    InstallOperation, RemoveOperation, UpdateOperation
)


def _pretty_operation(op):
    if isinstance(op, InstallOperation):
        s = "{} {}".format(op.package.name, op.package.version)
        return "InstallOperation(package={0})".format(s)
    if isinstance(op, RemoveOperation):
        s = "{} {}".format(op.package.name, op.package.version)
        return "RemoveOperation(package={0})".format(s)
    if isinstance(op, UpdateOperation):
        s0 = "{} {}".format(op.package.name, op.package.version)
        s1 = "{} {}".format(op.source.name, op.source.version)
        return "UpdateOperation(package={0}, source={1})".format(s0, s1)
    return str(op)


def _pkg_delta(operations, scenario_operations):
    pkg_delta = {}
    for p in operations:
        name = p.package.name
        pkg_delta.setdefault(name, [None, None])[0] = p
    for p in scenario_operations:
        name = p.package.name
        pkg_delta.setdefault(name, [None, None])[1] = p
    for n, v in list(six.iteritems(pkg_delta)):
        if _pretty_operation(v[0]) == _pretty_operation(v[1]):
            pkg_delta.pop(n)
    return pkg_delta


def _pretty_delta(pkg_delta):
    lines = []
    for k, v in sorted(pkg_delta.items()):
        lines.append(k)
        lines.append("  SOLVER  : {0}".format(_pretty_operation(v[0])))
        lines.append("  SCENARIO: {0}".format(_pretty_operation(v[1])))
        lines.append("")
    return '\n'.join(lines)


class ScenarioTestAssistant(object):

    def setUp(self):
        self.maxDiff = None

    def _check_solution(self, filename, prefer_installed=True):
        # Test that the solution described in the scenario file matches with
        # what the SAT solver computes.

        # Given
        scenario = Scenario.from_yaml(
            os.path.join(os.path.dirname(__file__), filename)
        )
        request = scenario.request

        # When
        pool = Pool(scenario.remote_repositories)
        pool.add_repository(scenario.installed_repository)

        solver = DependencySolver(
            pool, scenario.remote_repositories, scenario.installed_repository,
        )

        # Then
        try:
            transaction = solver.solve(request)
        except SatisfiabilityError as failure:
            self.assertFailureEqual(pool, failure, scenario)
        else:
            if scenario.failed:
                msg = "Solver unexpectedly succeeded, but {0}"
                self.fail(msg.format(scenario.failure['raw']))
            self.assertEqualOperations(transaction.operations,
                                       scenario.operations)
            if (scenario.pretty_operations or
                    transaction.operations != transaction.pretty_operations):
                self.assertEqualOperations(transaction.pretty_operations,
                                           scenario.pretty_operations)

    def assertFailureEqual(self, pool, failure, scenario):
        if not scenario.failed:
            msg = "Solver unexpectedly failed"
            if failure.unsat:
                reason = failure.unsat.to_string(pool=pool)
                msg += ":\n{0}".format(reason)
            self.fail(msg)

        result = [str(r) for r in failure.unsat.requirements]
        expected = scenario.failure['requirements']
        self.assertEqual(expected, result)

        result = failure.unsat.to_string(pool=pool)
        expected = scenario.failure['raw']
        self.assertMultiLineEqual(expected, result)

    def assertEqualOperations(self, operations, scenario_operations):
        pairs = zip(operations, scenario_operations)
        delta = _pretty_delta(_pkg_delta(operations, scenario_operations))
        for i, (left, right) in enumerate(pairs):
            left_s = "{0} {1}".format(left.package.name,
                                      left.package.version)
            right_s = "{0} {1}".format(right.package.name,
                                       right.package.version)
            if left_s != right_s:
                _pretty_delta(_pkg_delta(operations, scenario_operations))
                msg = "Item {0!r}: {1!r} vs {2!r}\n{3}".format(
                    i, left_s, right_s, delta)
                self.fail(msg)

            if not type(left) == type(right):
                msg = "Item {0!r} differ in kinds: {1!r} vs {2!r}\n{3}"
                self.fail(msg.format(i, left, right, delta))

        if len(operations) != len(scenario_operations):
            self.fail("Length of operations differ.\n{0}".format(delta))


class TestNoInstallSet(ScenarioTestAssistant, TestCase):

    def test_one_directly_implied_requirement(self):
        self._check_solution("one_directly_implied_requirement.yaml")

    def test_directly_implied_solution(self):
        self._check_solution("directly_implied_solution.yaml")

    def test_simple_numpy(self):
        self._check_solution("simple_numpy.yaml")

    def test_ipython(self):
        self._check_solution("ipython.yaml")

    def test_iris(self):
        self._check_solution("iris.yaml")

    def test_no_candidate(self):
        with self.assertRaises(NoPackageFound):
            self._check_solution("no_candidate.yaml")

    def test_three_way_conflict(self):
        self._check_solution("three_way_conflict.yaml")

    def test_conflicting_single_package_dependencies(self):
        self._check_solution("broken_metadata_self_conflict.yaml")

    def test_simple_provides(self):
        self._check_solution("simple_provides.yaml")

    def test_multiple_provides(self):
        self._check_solution("multiple_provides.yaml")

    def test_conflict_by_provides(self):
        self._check_solution("conflict_by_provides.yaml")


class TestInstallSet(ScenarioTestAssistant, TestCase):

    def test_simple_update_single(self):
        self._check_solution("simple_update_single.yaml")

    def test_update_single(self):
        self._check_solution("update_single.yaml")

    def test_no_prefer_installed(self):
        self._check_solution("no_prefer_installed.yaml",
                             prefer_installed=False)

    def test_update_all(self):
        self._check_solution("update_all.yaml")

    def test_update_all_conflict(self):
        self._check_solution("update_all_conflict.yaml")

    def test_update_all_noop(self):
        self._check_solution("update_all_noop.yaml")

    def test_ipython_upgrade(self):
        self._check_solution("ipython_upgrade.yaml")

    def test_simple_numpy(self):
        self._check_solution("simple_numpy_installed.yaml")

    def test_simple_numpy_installed_upgrade(self):
        self._check_solution("simple_numpy_installed_upgrade.yaml")

    def test_soft_update_with_deps(self):
        self._check_solution("soft_update_with_deps.yaml")

    def test_installed_pkg_missing_in_remote(self):
        self._check_solution("installed_pkg_missing_in_remote.yaml")

    def test_installed_pkg_missing_in_remote_not_requested(self):
        self._check_solution(
            "installed_pkg_missing_in_remote_not_requested.yaml"
        )

    def test_numpy_downgrade(self):
        self._check_solution("numpy_downgrade.yaml")

    def test_complex_numpy_downgrade(self):
        # Regression test for infinite loop in policy
        self._check_solution("complex_numpy_downgrade.yaml")

    def test_ipython(self):
        self._check_solution("ipython_with_installed.yaml")

    # This is currently handled using marked packages
    def test_blocked_upgrade(self):
        self._check_solution("preserve_marked_conflict.yaml")

    # This is currently handled using marked packages
    def test_blocked_downgrade(self):
        self._check_solution("preserve_marked_conflict_downgrade.yaml")

    def test_remove_no_reverse_dependencies(self):
        self._check_solution("simple_numpy_removed.yaml")

    def test_remove_reverse_dependencies(self):
        self._check_solution("remove_reverse_dependencies.yaml")

    def test_preserve_marked_packages(self):
        self._check_solution("preserve_marked.yaml")

    def test_remove_marked_packages(self):
        self._check_solution("remove_marked_package.yaml")

    def test_update_reverse_dependencies(self):
        self._check_solution(
            "update_reverse_dependencies.yaml",
            prefer_installed=True
        )

    def test_update_reverse_dependencies_no_prefer_installed(self):
        self._check_solution(
            "update_reverse_dependencies_no_prefer_installed.yaml",
            prefer_installed=False
        )

    def test_multiple_jobs(self):
        self._check_solution("multiple_jobs.yaml")

    def test_conflicts_simple(self):
        self._check_solution("explicit_conflict.yaml")

    def test_upgrade_conflict(self):
        self._check_solution("pil_pillow_upgrade_conflict.yaml")

    def test_downgrade_conflict(self):
        self._check_solution("pillow_pil_downgrade_conflict.yaml")

    def test_epd_full_conflict(self):
        self._check_solution("epd_full_conflict.yaml")

    def test_constraint_modifiers(self):
        self._check_solution('constraint_modifiers.yaml')



import abc

import six


def pkg_id_to_version(pool, package_id):
    return pool.id_to_package(package_id).version


class IPolicy(six.with_metaclass(abc.ABCMeta)):

    def __init__(self, *args):
        pass

    @abc.abstractmethod
    def add_requirements(self, package_ids):
        """ Submit packages to the policy for consideration.
        """

    @abc.abstractmethod
    def get_next_package_id(self, assignments, clauses):
        """ Returns a undecided variable (i.e. integer > 0) for the given sets
        of assignments and clauses.

        Parameters
        ----------
        assignments : AssignmentSet
            The current assignments of each literal. Keys are variables
            (integer > 0) and values are one of (True, False, None).
        clauses : List of Clause
            The collection of Clause objects to satisfy.
        """


class DefaultPolicy(IPolicy):

    def add_requirements(self, assignments):
        pass

    def get_next_package_id(self, assignments, *_):
        # Given a dictionary of partial assignments, get an undecided variable
        # to be decided next.
        undecided = (
            package_id for package_id, status in six.iteritems(assignments)
            if status is None
        )
        return next(undecided)



import unittest

from ..constraint_modifiers import ConstraintModifiers


class TestConstraintModifiers(unittest.TestCase):

    def test_initialization(self):
        # Given
        modifiers = ConstraintModifiers(allow_any='abc',
                                        allow_newer=('u', 'v'),
                                        allow_older=['x', 'y'])

        # Then
        self.assertEqual(modifiers.allow_any, set(('abc',)))
        self.assertEqual(modifiers.allow_newer, set(('u', 'v')))
        self.assertEqual(modifiers.allow_older, set(('x', 'y')))

    def test_asdict(self):
        # Given
        modifiers = ConstraintModifiers(allow_any=('c', 'b'),
                                        allow_newer=('z', 'a'),
                                        allow_older=['x', 'y'])

        # Then
        modifiers_dict = modifiers.asdict()
        expected_dict = {'allow_any': ['b', 'c'],
                         'allow_newer': ['a', 'z'],
                         'allow_older': ['x', 'y']}
        self.assertEqual(modifiers_dict, expected_dict)

    def test_update(self):
        # Given
        modifiers = ConstraintModifiers(allow_any=('a', 'b'))
        other_modifiers = ConstraintModifiers(allow_any='new',
                                              allow_newer=('u', 'v'))

        # Then
        modifiers.update(other_modifiers)
        self.assertEqual(modifiers.allow_any, set(('a', 'b', 'new')))
        self.assertEqual(modifiers.allow_newer, set(('u', 'v')))
        self.assertEqual(modifiers.allow_older, set())

    def test_remove(self):
        # Given
        modifiers = ConstraintModifiers(allow_any='x',
                                        allow_newer=('x', 'u', 'v'),
                                        allow_older=('x', 'y'))

        # When
        modifiers.remove(['x'])

        # Then
        self.assertEqual(modifiers.allow_any, set())
        self.assertEqual(modifiers.allow_newer, set(('u', 'v')))
        self.assertEqual(modifiers.allow_older, set(['y']))
        with self.assertRaises(TypeError):
            modifiers.remove("Don't pass a string")
        with self.assertRaises(TypeError):
            modifiers.remove(u"Don't pass unicode")
        with self.assertRaises(TypeError):
            modifiers.remove(b"Don't pass bytes")

    def test_targets(self):
        # Given
        modifiers = ConstraintModifiers(allow_any=('a', 'b', 'c'),
                                        allow_newer=('u', 'v'))

        # Then
        self.assertEqual(modifiers.targets, set(('a', 'b', 'c', 'u', 'v')))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division, print_function

import unittest
from textwrap import dedent

from simplesat.test_utils import pool_and_repository_from_packages

from ..graph import (
    package_lit_dependency_graph, toposort, transitive_neighbors
)


class TestGraph(unittest.TestCase):

    def test_transitive_neighbors(self):
        # Given
        graph = {
            0: [],
            1: [],
            2: [1],
            3: [0, 2],
            4: [1, 3],
            5: [0, 1, 2, 3, 4],
        }

        expected = {
            0: set(),
            1: set(),
            2: {1},
            3: {0, 1, 2},
            4: {0, 1, 2, 3},
            5: {0, 1, 2, 3, 4},
        }

        # When
        result = transitive_neighbors(graph)

        # Then
        self.assertEqual(result, expected)

        # Given
        graph = {
            0: [1],
            1: [0],
            2: [1],
            3: [3],
        }

        expected = {
            0: {0, 1},
            1: {0, 1},
            2: {0, 1},
            3: {3},
        }

        # When
        result = transitive_neighbors(graph)

        # Then
        self.assertEqual(result, expected)

    def test_toposort(self):
        # Given
        graph = {
            0: set(),
            1: set(),
            2: {1},
            3: {0, 2},
            4: {0},
            5: {1, 2},
            6: {3, 5},
        }

        expected = (
            {0, 1},
            {2, 4},
            {3, 5},
            {6}
        )

        # When
        result = tuple(toposort(graph))

        # Then
        self.assertEqual(expected, result)

    def test_toposort_cycle(self):
        # Given

        graph = {
            0: set(),
            1: set(), 2: {6},
            3: {0, 2},
            4: {0},
            5: {1, 2},
            6: {3, 5},
        }

        # Then
        with self.assertRaises(ValueError):
            tuple(toposort(graph))


class TestDependencyGraph(unittest.TestCase):

    def test_package_lit_dependency_graph(self):
        # Given
        package_def = dedent("""\
            A 0.0.0-1; depends (C ^= 0.0.0)
            A 3.0.0-1; depends (G ^= 1.0.0)
            B 0.0.0-1; depends (D ^= 0.0.0)
            B 1.0.0-1
            C 0.0.0-1
            D 0.0.0-1
            E 0.0.0-1; depends (A ^= 0.0.0)
            E 1.0.0-1; depends (I ^= 0.0.0, B ^= 1.0.0)
            F 0.0.0-1
            G 1.0.0-1
            I 0.0.0-1; depends (J ^= 0.0.0)
            J 0.0.0-1
            X 0.0.0-1; depends (A ^= 0.0.0, B ^= 0.0.0)
            X 1.0.0-1; depends (A ^= 3.0.0, B ^= 1.0.0)
            Y 0.0.0-1; depends (X ^= 0.0.0)
            Y 1.0.0-1
            Y 2.0.0-1; depends (F ^= 0.0.0)
            Y 3.0.0-1; depends (X ^= 1.0.0)
            Z 0.0.0-1; depends (E ^= 0.0.0)
            Z 1.0.0-1; depends (E ^= 1.0.0)
        """)

        expected = {
            1: {5},
            2: {10},
            3: {6},
            4: set(),
            5: set(),
            6: set(),
            7: {1},
            8: {11, 4},
            9: set(),
            10: set(),
            11: {12},
            12: set(),
            13: {1, 3},
            14: {2, 4},
            15: {13},
            16: set(),
            17: {9},
            18: {14},
            19: {7},
            20: {8},
        }

        # When
        pool, _ = pool_and_repository_from_packages(package_def)
        package_lits = pool.package_ids
        result = package_lit_dependency_graph(pool, package_lits)

        # Then
        self.assertEqual(expected, result)



import unittest
from textwrap import dedent

from simplesat import InstallRequirement, Repository
from simplesat.test_utils import packages_from_definition

from ..compute_dependencies import (compute_dependencies,
                                    compute_leaf_packages,
                                    compute_reverse_dependencies)


PACKAGE_DEF_0 = dedent("""\
    A 0.0.0-1; depends (B ^= 0.0.0)
    B 0.0.0-1; depends (D == 0.0.0-2)
    B 0.0.0-2; depends (D ^= 0.0.0)
    C 0.0.0-1; depends (E >= 1.0.0)
""")


PACKAGE_DEF_1 = dedent("""\
    D 0.0.0-2
    E 0.0.0-1
    E 1.0.0-1
    E 1.0.1-1
""")

PACKAGE_DEF_2 = dedent("""\
    B 0.0.0-1; depends (D == 0.0.0-2)
    C 0.0.0-1; depends (E >= 1.0.0)
""")


class TestComputeDependencies(unittest.TestCase):

    def setUp(self):
        repo_0 = Repository(packages_from_definition(PACKAGE_DEF_0))
        repo_1 = Repository(packages_from_definition(PACKAGE_DEF_1))
        self.repos = [repo_0, repo_1]

    def test_no_dependency(self):
        requirement = InstallRequirement._from_string('D == 0.0.0-2')
        expected_deps = set()
        deps = compute_dependencies(self.repos, requirement)
        self.assertEqual(deps, expected_deps)

    def test_simple_dependency(self):
        requirement = InstallRequirement._from_string('C *')
        expected_deps = packages_from_definition(
            """E 1.0.0-1
               E 1.0.1-1""")

        deps = compute_dependencies(self.repos, requirement)
        self.assertEqual(deps, set(expected_deps))

    def test_chained_requirements(self):
        requirement = InstallRequirement._from_string('A ^= 0.0.0')
        expected_deps = packages_from_definition(
            """B 0.0.0-1; depends (D == 0.0.0-2)
            B 0.0.0-2; depends (D ^= 0.0.0) """
        )

        deps = compute_dependencies(self.repos, requirement)
        self.assertEqual(deps, set(expected_deps))

    def test_chained_requirements_transitive(self):
        requirement = InstallRequirement._from_string('A ^= 0.0.0')
        expected_deps = packages_from_definition(
            """B 0.0.0-1; depends (D == 0.0.0-2)
            B 0.0.0-2; depends (D ^= 0.0.0)
            D 0.0.0-2 """
        )

        deps = compute_dependencies(self.repos, requirement, transitive=True)
        self.assertEqual(deps, set(expected_deps))


class TestComputeReverseDependencies(unittest.TestCase):

    def setUp(self):
        repo_0 = Repository(packages_from_definition(PACKAGE_DEF_0))
        repo_1 = Repository(packages_from_definition(PACKAGE_DEF_1))
        self.repos = [repo_0, repo_1]

    def test_no_dependency(self):
        requirement = InstallRequirement._from_string('A ^= 0.0.0')

        deps = compute_reverse_dependencies(self.repos, requirement)
        self.assertEqual(deps, set())

    def test_simple_dependency(self):
        requirement = InstallRequirement._from_string('E *')
        expected_deps = packages_from_definition(
            'C 0.0.0-1; depends (E >= 1.0.0)'
        )

        deps = compute_reverse_dependencies(self.repos, requirement)
        self.assertEqual(deps, set(expected_deps))

    def test_chained_dependencies(self):
        requirement = InstallRequirement._from_string('D ^= 0.0.0')
        expected_deps = packages_from_definition(
            """B 0.0.0-1; depends (D == 0.0.0-2)
            B 0.0.0-2; depends (D ^= 0.0.0)"""
        )
        deps = compute_reverse_dependencies(self.repos, requirement)
        self.assertEqual(deps, set(expected_deps))

    def test_chained_dependencies_transitive(self):
        requirement = InstallRequirement._from_string('D ^= 0.0.0')
        expected_deps = packages_from_definition(
            """A 0.0.0-1; depends (B ^= 0.0.0)
            B 0.0.0-1; depends (D == 0.0.0-2)
            B 0.0.0-2; depends (D ^= 0.0.0)"""
        )
        deps = compute_reverse_dependencies(self.repos, requirement,
                                            transitive=True)
        self.assertEqual(deps, set(expected_deps))


class TestComputeLeafPackages(unittest.TestCase):

    def setUp(self):
        repo_0 = Repository(packages_from_definition(PACKAGE_DEF_0))
        repo_1 = Repository(packages_from_definition(PACKAGE_DEF_1))
        repo_2 = Repository(packages_from_definition(PACKAGE_DEF_2))
        self.repos = [repo_0, repo_1, repo_2]

    def test_simple(self):
        expected_leaf_packages = packages_from_definition(
            """A 0.0.0-1; depends (B ^= 0.0.0)
            C 0.0.0-1; depends (E >= 1.0.0)
            E 0.0.0-1 """
        )
        leaf_packages = compute_leaf_packages(self.repos)

        self.assertEqual(leaf_packages, set(expected_leaf_packages))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division, print_function

import unittest
from textwrap import dedent

from simplesat.test_utils import pool_and_repository_from_packages

from ..transaction import (
    InstallOperation, RemoveOperation, Transaction, UpdateOperation
)


PACKAGE_DEF = dedent("""\
    A 0.0.0-1; depends (C ^= 0.0.0)
    A 3.0.0-1; depends (G ^= 1.0.0)
    B 0.0.0-1; depends (D ^= 0.0.0)
    B 1.0.0-1
    C 0.0.0-1
    D 0.0.0-1
    E 0.0.0-1; depends (A ^= 0.0.0)
    E 1.0.0-1; depends (I ^= 0.0.0, B ^= 1.0.0)
    F 0.0.0-1
    G 1.0.0-1
    I 0.0.0-1; depends (J ^= 0.0.0)
    J 0.0.0-1
    X 0.0.0-1; depends (A ^= 0.0.0, B ^= 0.0.0)
    X 1.0.0-1; depends (A ^= 3.0.0, B ^= 1.0.0)
    Y 0.0.0-1; depends (X ^= 0.0.0)
    Y 1.0.0-1
    Y 2.0.0-1; depends (F ^= 0.0.0)
    Y 3.0.0-1; depends (X ^= 1.0.0)
    Z 0.0.0-1; depends (E ^= 0.0.0)
    Z 1.0.0-1; depends (E ^= 1.0.0)
""")


class TestTransaction(unittest.TestCase):

    maxDiff = None

    def setUp(self):
        self.pool, self.repo = pool_and_repository_from_packages(PACKAGE_DEF)

    def test_operations(self):

        # Given
        installed = {1, 3, 5, 6, 7, 9, 10, 11, 12, 13, 15, 19}
        wanted = {20, 8, 11, 12, 4}
        decisions = wanted.union(-i for i in installed if i not in wanted)
        to_install = (4, 8, 20)
        to_remove = (19, 15, 13, 7, 3, 1, 10, 9, 6, 5)

        # When
        installs = [InstallOperation(self.pool.id_to_package(i))
                    for i in to_install]
        removals = [RemoveOperation(self.pool.id_to_package(i))
                    for i in to_remove]
        transaction = Transaction(self.pool, decisions, installed)
        result = transaction.operations
        expected = removals + installs

        # Then
        self.assertListEqual(expected, result)

    def test_pretty_operations(self):

        # Given
        installed = {1, 3, 5, 6, 7, 9, 10, 11, 12, 13, 15}
        wanted = {20, 8, 11, 12, 4}
        decisions = wanted.union(-i for i in installed if i not in wanted)
        to_install = (20,)
        to_update = ((3, 4), (7, 8))
        to_remove = (15, 13, 1, 10, 9, 6, 5)

        # When
        installs = [InstallOperation(self.pool.id_to_package(i))
                    for i in to_install]
        updates = [UpdateOperation(self.pool.id_to_package(new),
                                   self.pool.id_to_package(old))
                   for old, new in to_update]
        removals = [RemoveOperation(self.pool.id_to_package(i))
                    for i in to_remove]
        transaction = Transaction(self.pool, decisions, installed)
        result = transaction.pretty_operations
        expected = removals + updates + installs

        # Then
        self.assertListEqual(expected, result)



import abc

import six

from okonomiyaki.versions import EnpkgVersion


class IConstraint(six.with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def matches(self, version):
        """Returns True if the version object matches this constraint.

        Parameters
        ==========
        version : Version
            A version object
        """


class Any(IConstraint):
    def matches(self, candidate_version):
        return True

    def __str__(self):
        return ""

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.__class__)


class IVersionConstraint(IConstraint):
    def __init__(self, version):
        self.version = version

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__ and
            self.version == other.version
        )

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.version)


class Equal(IVersionConstraint):
    def __str__(self):
        return "== " + str(self.version)

    def matches(self, candidate_version):
        return self.version == candidate_version


class Not(IVersionConstraint):
    def __str__(self):
        return "!= " + str(self.version)

    def matches(self, candidate_version):
        return self.version != candidate_version


class GEQ(IVersionConstraint):
    def __str__(self):
        return ">= " + str(self.version)

    def matches(self, candidate_version):
        return candidate_version >= self.version


class GT(IVersionConstraint):
    def __str__(self):
        return "> " + str(self.version)

    def matches(self, candidate_version):
        return candidate_version > self.version


class LEQ(IVersionConstraint):
    def __str__(self):
        return "<= " + str(self.version)

    def matches(self, candidate_version):
        return candidate_version <= self.version


class LT(IVersionConstraint):
    def __str__(self):
        return "< " + str(self.version)

    def matches(self, candidate_version):
        return candidate_version < self.version


class EnpkgUpstreamMatch(IVersionConstraint):
    def __init__(self, version):
        if not isinstance(version, EnpkgVersion):
            raise ValueError(
                "Invalid type for {0!r}: {1!r}".format(version, type(version))
            )
        super(EnpkgUpstreamMatch, self).__init__(version)

    def __str__(self):
        return "^= " + str(self.version.upstream)

    def matches(self, candidate_version):
        return candidate_version.upstream == self.version.upstream



import re

import six

from okonomiyaki.versions import EnpkgVersion

from simplesat.errors import (
    InvalidConstraint, InvalidDependencyString, SolverException
)

from .kinds import Any, Equal
from .multi import MultiConstraints
from .parser import (
    _RawConstraintsParser, _RawRequirementParser,
    _DISTRIBUTION_NAME_R, _VERSION_R
)


_FULL_PACKAGE_RC = re.compile("""\
        (?P<name>{})
        (?:-|\s+)
        (?P<version>{})
        $""".format(_DISTRIBUTION_NAME_R, _VERSION_R),
    re.VERBOSE)


def parse_package_full_name(full_name):
    """
    Parse a package full name (e.g. 'numpy-1.6.0-1') into a (name,
    version_string) pair.
    """
    m = _FULL_PACKAGE_RC.match(full_name)
    if m:
        return m.group("name"), m.group("version")
    else:
        msg = "Invalid package full name {0!r}".format(full_name)
        raise SolverException(msg)


def _first(iterable):
    return six.next(iter(iterable))


class Requirement(object):
    """Requirements instances represent a 'package requirement', that is a
    package + version constraints.

    Arguments
    ---------
    name: str
        PackageInfo name
    specs: seq
        Sequence of constraints
    """

    @classmethod
    def from_constraints(cls, constraint_tuple):
        """ Return a Requirement object from a PackageMetadata constraint
        tuple.

        Parameters
        ----------
        constraints : constraints tuple
            A 2-tuple of constraints where the first element is the
            distribution name, and the second is a tuple of tuple of string,
            representing a disjuntion of conjunctions of version ranges, e.g.
            ``('nose', (('< 1.4', '>= 1.3'),))``.


            >>> Requirement.from_constraints((
            ...     'nose', ( # disjunction
            ...         ('< 1.4', '>= 1.3'),  # conjunction
            ...     )
            ... ))
            Requirement('nose < 1.4-0, >= 1.3-0')

        Returns
        -------
        Requirement
            A Requirement that matches the given constraints.

        Raises
        ------
        InvalidConstraint
            If there is more than one conjunction. In less formal terms, we do
            not currently support the OR operator.
        InvalidConstraint
            If the constraint tuple has the wrong shape.
        """
        try:
            name, disjunction = constraint_tuple
        except ValueError:
            msg = "Invalid constraint tuple: {}"
            raise InvalidConstraint(msg.format(constraint_tuple))

        if len(disjunction) > 1:
            msg = "Disjunction (OR) is not yet supported in constraints: {}"
            raise InvalidConstraint(msg.format(disjunction))

        parse = _RawConstraintsParser().parse

        def insert_star(conjunction):
            return conjunction if len(conjunction) > 0 else ("*",)

        constraints = tuple(
            constraint
            for conjunction in disjunction
            for constraint_str in insert_star(conjunction)
            for constraint in parse(constraint_str, EnpkgVersion.from_string))

        return cls(name, constraints)

    def to_constraints(self):
        """ Return a constraint tuple as described by :meth:`from_constraints`.
        """
        name = self.name

        parts = []
        for constraint in self._constraints._constraints:
            if not isinstance(constraint, Any):
                parts.append(str(constraint))
            else:
                parts.append('*')
        return (name, (tuple(parts),))

    @classmethod
    def _from_string(cls, string,
                     version_factory=EnpkgVersion.from_string):
        """ Creates a requirement from a requirement string.

        Parameters
        ----------
        string : str
            The requirement string, e.g. 'MKL >= 10.3, MKL < 11.0'
        version_factory : callable, optional
            A function from version strings to version objects.

        Returns
        -------
        Requirement
            A requirement matching the `string`.

        Raises
        ------
        InvalidDependencyString
            If the string refers to more than one package.
        """
        parser = _RawRequirementParser()
        named_constraints = parser.parse(string, version_factory)
        if len(named_constraints) > 1:
            names = named_constraints.keys()
            msg = "Multiple package name for constraint: {0!r}".format(names)
            raise InvalidDependencyString(msg)
        assert len(named_constraints) > 0
        name = _first(named_constraints.keys())
        return cls(name, named_constraints[name])

    @classmethod
    def from_package_string(cls, package_string,
                            version_factory=EnpkgVersion.from_string):
        """ Creates a requirement from a package full version.

        Parameters
        ----------
        package_string : str
            The package string, e.g. 'numpy-1.8.1-1'
        version_factory : callable, optional
            A function from version strings to version objects.

        Returns
        -------
        Requirement
            A requirement matching the exact package and version in
            `package_string`.
        """
        name, version_string = parse_package_full_name(package_string)
        version = version_factory(version_string)
        return cls(name, (Equal(version),))

    def __init__(self, name, constraints=None):
        self.name = name
        self._constraints = MultiConstraints(constraints)

    def matches(self, version_candidate):
        """ Returns True if the given version matches this set of
        requirements, False otherwise.

        Parameters
        ----------
        version_candidate : obj
            A valid version object (must match the version factory of the
            requirement instance).
        """
        return self._constraints.matches(version_candidate)

    def __eq__(self, other):
        return (self.name == other.name and
                self._constraints == other._constraints)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.name, self._constraints))

    def __str__(self):
        name, constraints = self.to_constraints()
        parts = [
            constraint
            for conjunction in constraints
            for constraint in conjunction
            if constraint != '*'
        ]

        if len(parts) == 0:
            return self.name
        else:
            return self.name + " " + ", ".join(parts)

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self)

    @property
    def has_any_version_constraint(self):
        """ True if there is any version constraint."""
        constraints = self._constraints._constraints
        if len(constraints) == 0:
            return False
        elif len(constraints) == 1:
            constraint = six.next(iter(constraints))
            if isinstance(constraint, Any):
                return False
        return True


class InstallRequirement(Requirement):
    """ A Requirement that describes packages to be installed. """


class ConflictRequirement(Requirement):
    """ A Requirement that describes packages which must not be installed. """



"""Run the simple SAT solver on the binary case of van der waerden's problem.

This problem asks for the smallest number n so that a binary number of n digits
that contains either j digits 0 or k digits 1, for given integers j and k.

"""
from __future__ import division


def _van_der_waerden_helper(j, n, sign):
    clause_data = []
    max_d = (n - 1) // (j - 1) + 1
    for d in range(1, max_d + 1):
        for i in range(1, n - (j - 1) * d + 1):
            clause = [sign * (i + p * d) for p in range(0, j)]
            clause_data.append(clause)
    return clause_data


def van_der_waerden(j, k, n):
    """Generate clauses for the van der Waerden problem.
    """
    return (
        _van_der_waerden_helper(j, n, +1) +
        _van_der_waerden_helper(k, n, -1)
    )



import collections
import itertools

import attr
import six

from simplesat.constraints import ConstraintModifiers
from simplesat.constraints.requirement import InstallRequirement
from simplesat.errors import (
    NoPackageFound, SatisfiabilityError, SatisfiabilityErrorWithHint,
    UnexpectedlySatisfiable)
from simplesat.pool import Pool
from simplesat.repository import Repository
from simplesat.request import JobType, Request
from simplesat.rules_generator import RulesGenerator
from simplesat.sat.policy import InstalledFirstPolicy
from simplesat.sat import MiniSATSolver
from simplesat.transaction import Transaction, InstallOperation
from simplesat.utils import timed_context, connected_nodes


def requirements_from_packages(packages):
    """
    Return a list of requirements, one to match each package in `packages`.

    Parameters
    ----------
    packages : iterable of PackageMetadata
        The packages for which to generate requirements.

    Returns
    -------
    tuple of Requirement
        The matching requirements.
    """
    R = InstallRequirement.from_package_string
    return tuple(R("{0.name}-{0.version}".format(package))
                 for package in packages)


def packages_from_requirements(packages, requirements, modifiers=None):
    """
    Return a new tuple that only contains packages explicitly mentioned
    in the requirements.

    Parameters
    ----------
    packages : iterable of PackageMetadata
        The packages available for inclusion in the result.
    requirements : list of Requirement
        The requirements used to identify relevant packages. All packages that
        satisfy any of the requirements will be included.
    modifiers : ConstraintModifiers, optional
        If not None, modify requirements before resolving packages.

    Returns
    -------
    Tuple of PackageMetadata
        A tuple containing the relevant packages.
    """
    pool = Pool((Repository(packages),), modifiers=modifiers)
    listed_packages = set()
    for requirement in requirements:
        listed_packages.update(pool.what_provides(requirement))
    return tuple(sorted(listed_packages, key=lambda p: p._key))


def packages_are_consistent(packages, modifiers=None):
    """
    Return True if all packages can be installed together.

    .. Note::
        This will return `False` if more than one version of a package is
        present because we only permit one at a time.

    Parameters
    ----------
    packages : iterable of PackageMetadata
        The packages to check for consistency.
    modifiers : ConstraintModifiers, optional
        If not None, modify requirements before resolving packages.

    Returns
    -------
    bool
        True if every package in `packages` can be installed
        simultaneously, otherwise False.
    """
    requirements = requirements_from_packages(packages)
    return requirements_are_satisfiable(
        packages, requirements, modifiers=modifiers)


def requirements_are_complete(packages, requirements, modifiers=None):
    """
    Return True if `requirements` includes all required transitive
    dependencies. I.e. it will report whether all the packages that are needed
    are explicitly required.

    Parameters
    ----------
    packages : iterable of PackageMetadata
        The packages available to draw from when satisfying requirements.
    requirements : iterable of Requirement
        The requirements used to identify relevent packages.
    modifiers : ConstraintModifiers, optional
        If not None, modify requirements before resolving packages.

    Returns
    -------
    bool
        True if the requirements specify all necessary packages.
    """
    packages = packages_from_requirements(
        packages, requirements, modifiers=modifiers)
    return requirements_are_satisfiable(
        packages, requirements, modifiers=modifiers)


def requirements_are_satisfiable(packages, requirements, modifiers=None):
    """ Determine if the `requirements` can be satisfied together.

    Parameters
    ----------
    packages : iterable of PackageMetadata
        The packages available to draw from when satisfying requirements.
    requirements : list of Requirement
        The requirements used to identify relevent packages.
    modifiers : ConstraintModifiers, optional
        If not None, modify requirements before resolving packages.

    Returns
    -------
    bool
        Return True if the `requirements` can be satisfied by the `packages`.
    """
    modifiers = modifiers or ConstraintModifiers()
    Result = collections.namedtuple("Result", "is_satisfiable message")

    request = Request(modifiers=modifiers)
    for requirement in requirements:
        request.install(requirement)
    repositories = (Repository(packages),)
    pool = Pool(repositories, modifiers=modifiers)

    try:
        DependencySolver(pool, repositories, []).solve(request)
        return Result(is_satisfiable=True, message="")
    except SatisfiabilityError as e:
        return Result(
            is_satisfiable=False, message=e.unsat.to_string(pool=pool)
        )


def satisfy_requirements(packages, requirements, modifiers=None):
    """ Find a collection of packages that satisfy the requirements.

    Parameters
    ----------
    packages : iterable of PackageMetadata
        The packages available to draw from when satisfying requirements.
    requirements : list of Requirement
        The requirements used to identify relevent packages.
    modifiers : ConstraintModifiers, optional
        If not None, modify requirements before resolving packages.

    Returns
    -------
    tuple of PackageMetadata
        Return a tuple of packages that together satisfy all of the
        `requirements`. The packages are in topological order.

    Raises
    ------
    SatisfiabilityError
        If the `requirements` cannot be satisfied using the `packages`.
    """
    request = Request(modifiers=modifiers)
    for requirement in requirements:
        request.install(requirement)
    repositories = (Repository(packages),)
    pool = Pool(repositories, modifiers=modifiers)
    transaction = DependencySolver(pool, repositories, []).solve(request)
    msg = ("""
        Unexpected operation in the transaction. This should never occur.
        Something in simplesat is broken.
        {!r}""")
    for op in transaction.operations:
        # Our installed repository was empty so everything should be an install
        # operation
        assert isinstance(op, InstallOperation), msg.format(op)
    packages = tuple(op.package for op in transaction.operations)
    return packages


def simplify_requirements(packages, requirements):
    """ Return a reduced, but equivalent set of requirements.

    Parameters
    ----------
    packages : iterable of PackageMetadata
        The packages available to draw from when satisfying requirements.
    requirements : list of Requirement
        The requirements used to identify relevent packages.

    Returns
    -------
    tuple of Requirement
        The reduced requirements.
    """

    needed_packages = packages_from_requirements(packages, requirements)
    pool = Pool([Repository(packages)])
    R = InstallRequirement.from_constraints
    dependencies = set(itertools.chain.from_iterable(
        pool.what_provides(R(con))
        for package in needed_packages
        for con in package.install_requires
    ))
    simple_requirements = requirements_from_packages(
        package
        for package in needed_packages
        if package not in dependencies
    )
    return simple_requirements


def minimal_unsatisfiable_subset(clauses, callback):
    """
    Given a set of clauses, find a minimal unsatisfiable subset (an
    unsatisfiable core)

    A set is a minimal unsatisfiable subset if no proper subset is
    unsatisfiable.  A set of clauses may have many minimal unsatisfiable
    subsets of different sizes.

    Parameters
    ----------
    clauses : seq
        Iterator over a set of clauses. Clause should follow the format
        expected by callback
    callback : callable
        A function of the form f: <clauses subset> -> bool, returning True if
        the given subset of clauses is satisfiable, False otherwise. It should
        follow the constraint A <= B iff (f(B) <= f(A)), with False < True.

    Note
    ----
    Function adapted from conda (conda.logic as of 4.0.5), itself implemented
    from http://www.slideshare.net/pvcpvc9/lecture17-31382688. The conda
    function mentions the following:

    We do a binary search on the clauses by splitting them in halves A and B.
    If A or B is UNSAT, we use that and repeat. Otherwise, we recursively check
    A, but each time we do a sat query, we include B, until we have a minimal
    subset A* of A such that A* U B is UNSAT. Then we find a minimal subset B*
    of B such that A* U B* is UNSAT. Then A* U B* will be a minimal
    unsatisfiable subset of the original set of clauses.

    Proof: If some proper subset S of A* U B* is UNSAT, then there is some
    clause c in A* U B* not in S. If c is in A*, then that means (A* - {c}) U
    B* is UNSAT, and hence (A* - {c}) U B is UNSAT, since it is a superset,
    which contradicts A* being the minimal subset of A with such property.
    Similarly, if c is in B, then A* U (B* - {c}) is UNSAT, but B* - {c} is a
    strict subset of B*, contradicting B* being the minimal subset of B with
    this property.
    """
    clauses = tuple(clauses)
    if callback(clauses):
        raise UnexpectedlySatisfiable()

    def split(s):
        """
        Split s into two roughly equal parts
        """
        s = tuple(s)
        size = len(s)
        return s[:size//2], s[size//2:]

    def minimal_unsat(clauses, include=()):
        """
        Return a minimal subset A of clauses such that A + include is
        unsatisfiable.

        Implicitly assumes that clauses + include is unsatisfiable.
        """
        # Base case: Since clauses + include is implicitly assumed to be
        # unsatisfiable, if clauses has only one element, it must be its own
        # minimal subset
        if len(clauses) == 1:
            return clauses

        A, B = split(clauses)

        # If one half is unsatisfiable (with include), we can discard the
        # other half.
        if not callback(A + include):
            return minimal_unsat(A, include)
        if not callback(B + include):
            return minimal_unsat(B, include)

        Astar = minimal_unsat(A, B + include)
        Bstar = minimal_unsat(B, Astar + include)
        return Astar + Bstar

    return minimal_unsat(clauses)


class DependencySolver(object):

    """
    Top-level class for resolving a package management scenario.

    The solver is configured at construction time with packages repositories
    and a :class:`Policy` and exposes an API for computing a
    :class:`Transaction` that describes what to do.

    Parameters
    ----------
    pool : Pool
        Pool against which to resolve Requirements.
    remote_repositories : list of Repository
        Repositories containing package available for installation.
    installed_repository : Repository
        Repository containing the packages which are currently installed.
    use_pruning : bool, optional
        When True, attempt to prune package operations that are not strictly
        necessary for meeting the requirements. Without this, packages whose
        assignments have changed as an artefact of the search process, but
        which are not needed for the solution will be modified.

        A typical example might be the installation of a dependency for a
        package that was proposed but later backtracked away.
    strict : bool, optional
        When true, behave more harshly when dealing with broken packages. INFO
        level log messages become WARNINGs and missing dependencies become
        errors rather than causing the package to be ignored.


    >>> from simplesat.constraints.package_parser import \\
    ...     pretty_string_to_package as P
    >>> numpy1921 = P('numpy 1.9.2-1; depends (MKL 10.2-1)')
    >>> mkl = P('MKL 10.3-1')
    >>> installed_repository = Repository([mkl])
    >>> remote_repository = Repository([mkl, numpy1921])
    >>> request = Request()
    >>> request.install(Requirement.from_string('numpy >= 1.9'))
    >>> request.allow_newer('MKL')
    >>> pool = Pool([installed_repo] + remote_repos)
    >>> pool.modifiers = request.modifiers
    >>> solver = DependencySolver(pool, remote_repos, installed_repo)
    >>> transaction = solver.solve(request)
    """

    def __init__(self, pool, remote_repositories, installed_repository,
                 use_pruning=True, strict=False):
        self._pool = pool
        self._installed_repository = installed_repository

        if isinstance(remote_repositories, Repository):
            raise ValueError(
                u"remote_repositories should be a sequence of "
                 "repositories, not a repository"
            )

        self._remote_repositories = remote_repositories
        self._last_rules_time = timed_context("Generate Rules")
        self._last_solver_init_time = timed_context("Solver Init")
        self._last_solve_time = timed_context("SAT Solve")

        self.strict = strict
        self.use_pruning = use_pruning

    def solve(self, request):
        """Given a request return a Transaction that would satisfy it.

        Parameters
        ----------
        request : Request
            The request that should be satisifed.

        Returns
        -------
        Transaction
            The operations to apply to resolve the `request`.

        Raises
        ------
        SatisfiabilityError
            If no resolution is found.
        """
        request = _convert_upgrade_request_if_needed(
            request, self._remote_repositories, self._installed_repository
        )

        modifiers = request.modifiers
        self._pool.modifiers = modifiers if modifiers.targets else None
        with self._last_rules_time:
            init_rules_and_policy = self._create_rules_and_initialize_policy
            requirement_ids, rules, policy = init_rules_and_policy(
                request
            )
        with self._last_solver_init_time:
            sat_solver = MiniSATSolver.from_rules(rules, policy)
        with self._last_solve_time:
            solution = sat_solver.search()
        solution_ids = _solution_to_ids(solution)

        installed_package_ids = set(
            self._pool.package_id(p)
            for p in self._installed_repository
        )

        if self.use_pruning:
            root_ids = installed_package_ids.union(requirement_ids)
            solution_ids = _connected_packages(
                solution_ids, root_ids, self._pool
            )

        return Transaction(self._pool, solution_ids, installed_package_ids)

    def solve_with_hint(self, request):
        """Given a request return a Transaction that would satisfy it.

        If the solver cannot find a solution, it will raise a
        SatisfiabilityErrorWithHint exception, that contains enough information
        to give a human-readable error message.

        Parameters
        ----------
        request : Request
            The request that should be satisifed.

        Returns
        -------
        Transaction
            The operations to apply to resolve the `request`.

        Raises
        ------
        SatisfiabilityErrorWithHint
            If no resolution is found.
        """
        try:
            return self.solve(request)
        except SatisfiabilityError as exc:
            # XXX: we recompute the upgrade request here so that we get the
            # corresponding jobs to be used by minimal_unsatisfiable_subset.
            request = _convert_upgrade_request_if_needed(
                request, self._remote_repositories, self._installed_repository
            )

            def callback(jobs):
                callback_request = Request(modifiers=request.modifiers)
                for job in jobs:
                    callback_request.jobs.append(job)
                try:
                    self.solve(callback_request)
                    return True
                except SatisfiabilityError:
                    return False
            conflicting_jobs = minimal_unsatisfiable_subset(request.jobs, callback)
            raise SatisfiabilityErrorWithHint(exc.unsat, conflicting_jobs)

    def _create_rules_and_initialize_policy(self, request):
        pool = self._pool
        installed_repository = self._installed_repository

        all_requirement_ids = []

        soft_update_packages = set()
        for job in request.jobs:
            assert job.kind in (
                JobType.install, JobType.remove, JobType.hard_update,
                JobType.soft_update
            ), 'Unknown job kind: {}'.format(job.kind)

            requirement = job.requirement

            providers = tuple(pool.what_provides(
                requirement, use_modifiers=False))
            if len(providers) == 0:
                raise NoPackageFound(requirement, str(requirement))

            if job.kind == JobType.hard_update:
                # An update request *must* install the latest package version
                def key(package):
                    return (package.version, package in installed_repository)
                providers = [max(providers, key=key)]
            elif job.kind == JobType.soft_update:
                soft_update_packages.update((pkg for pkg in providers
                                            if pkg in installed_repository))

            all_requirement_ids.extend(pool.package_id(p) for p in providers)

        installed_package_ids = collections.OrderedDict()
        for package in installed_repository:
            package_id = pool.package_id(package)
            installed_package_ids[package_id] = package

        # Prefer the installed versions of all packages
        policy = InstalledFirstPolicy(
            pool, installed_repository,
            ignore_installed_packages=soft_update_packages)
        policy.add_requirements(all_requirement_ids)

        rules_generator = RulesGenerator(
            pool, request, installed_package_ids=installed_package_ids,
            strict=self.strict)

        return all_requirement_ids, list(rules_generator.iter_rules()), policy


def _convert_upgrade_request_if_needed(request, remote_repositories,
                                       installed_repository):

    if len(request.jobs) == 1 and request.jobs[0].kind == JobType.upgrade:
        upgrade_request = attr.assoc(request, jobs=[])
        remote_repository = Repository()
        for repository in remote_repositories:
            remote_repository.update(repository)

        latest_packages = []
        for package in installed_repository:
            candidates = remote_repository.find_packages(package.name)
            # candidates may be empty (e.g. when repository configuration
            # changed, and an installed package is coming from a repository not
            # configured in the remote list)
            if len(candidates) > 0:
                latest_packages.append(candidates[-1])

        for p in latest_packages:
            upgrade_request.install(
                InstallRequirement._from_string("{} == {}".format(p.name, p.version))
            )

        return upgrade_request
    else:
        return request


def _connected_packages(solution, root_ids, pool):
    """ Return packages in `solution` which are associated with `root_ids`. """

    # Our strategy is as follows:
    # ... -> pkg.install_requires -> pkg names -> ids -> _id_to_package -> ...

    def get_names(pkg_id):
        provides = pool.id_to_package(abs(pkg_id)).provides
        return tuple(name for name, _ in provides)

    root_names = {name for pkg_id in root_ids for name in get_names(pkg_id)}

    solution_name_to_id = {
        name: pkg_id
        for pkg_id in solution
        for name in get_names(pkg_id)
        if pkg_id > 0
    }

    solution_root_ids = set(
        pkg_id for name, pkg_id in six.iteritems(solution_name_to_id)
        if name in root_names
    )

    def neighborfunc(pkg_id):
        """ Given a pkg id, return the pkg ids of the immediate dependencies
        that appeared in our solution. """
        constraints = pool.id_to_package(pkg_id).install_requires
        neighbors = set(solution_name_to_id[name] for name, _ in constraints)
        return neighbors

    # Each package can root its own independent graph, so we must start at
    # each one individually
    should_include = set()
    for pkg_id in solution_root_ids:
        # We pass in `should_include` to avoid re-walking a subgraph
        nodes = connected_nodes(pkg_id, neighborfunc, should_include)
        should_include.update(nodes)
    assert should_include.issuperset(solution_root_ids)

    # In addition to all updates and additions to root ids, we must also keep
    # all packages newly *excluded* from root_ids
    connected = should_include.union(s for s in solution if abs(s) in root_ids)
    return connected


def _solution_to_ids(solution):
    # Return solution as list of signed integers.
    ids = (pkg_id if value else -pkg_id
           for pkg_id, value in six.iteritems(solution))
    return sorted(ids, key=lambda lit: abs(lit))



"""
A SAT solver which follows the Minisat approach.

"""
from __future__ import absolute_import

from collections import defaultdict, deque, OrderedDict
import itertools

from six.moves import range

from simplesat.errors import SatisfiabilityError
from .assignment_set import AssignmentSet
from .clause import Clause
from .policy import DefaultPolicy
from simplesat.utils import timed_context
from simplesat.utils.graph import breadth_first_search


class UNSAT(object):

    """An unsatisfiable set of boolean clauses."""

    def __init__(self, conflict_clause, learned_clause, trails,
                 assigning_clauses):
        """
        Create a new UNSAT object.

        Parameters
        ----------
        conflict : Clause
            The clause which has been found to be unsatisfiable.
        learned_clause : Clause
            The implied change needed to satisfiy the conflict.

            This clause is always a learned clause and should always have
            exactly one literal associated with it; the conflicting assignment.
        trails : dict
            A mapping from clauses to the trail of clauses that generated them.
            Only learned clauses should have trails of non-zero length
        """

        self._conflict_clause = conflict_clause
        self._learned_clause = learned_clause
        self._clause_trails = trails
        self._assigning_clauses = assigning_clauses

        # A flattened version of `self._clause_trails`
        self._flat_clause_trails = {}

        # A mapping from clauses to the requirements that generated them
        self._clause_requirements = {}

        # The a list of lists representing "problems". These are the clauses
        # that we'll use to construct our paths, one per problem.
        self._conflict_details = []
        self._conflict_paths = []

        self._find_requirement_time = None
        with timed_context("Find Requirements") as self._find_requirement_time:
            # We'll look at the chain of clauses that led us to assign the
            # original value and then the chain of clauses that led us to want
            # to assign the opposite value
            assert len(learned_clause.lits) == 1
            self._implicand = -learned_clause[0]
            self._implicand_clause = assigning_clauses[abs(self._implicand)]
            assert self._implicand_clause is not None

            # The clauses that led us to our first assignment
            implicand_req_clauses = self.clause_requirements(
                self._implicand_clause)
            # The clauses we used to learn that the first assignment is invalid
            learned_req_clauses = self.clause_requirements(learned_clause)
            # The clause we were on when we discovered the conflict.
            conflicting_req_clauses = self.clause_requirements(conflict_clause)
            self._conflict_details.extend(
                implicand_req_clauses +
                learned_req_clauses +
                conflicting_req_clauses)

        # Figure out the paths for the conflict(s) we've found
        # These paths are a minimal series of related clauses that describe the
        # conflict(s).
        clauses = self._conflict_details
        end_clauses = self._end_clauses(clauses, implicand=self._implicand)
        paths = self._find_conflict_paths(end_clauses, clauses)
        self._conflict_paths.extend(paths)

    def _key(self, clause):
        return tuple(sorted(l for l in clause.lits))

    def _find_conflict_paths(self, end_clauses, relevant_clauses):
        """ Return a tuple of paths representing conflicts between a set of
        clauses.

            See https://github.com/enthought/sat-solver/wiki/Unsatisfiability-Error-Messages
            for discussion about how best to implement this.
        """  # noqa
        # It's expensive to figure out which clauses are neighbors. This dict
        # maps ids to clauses containing that id. We can do this lookup for
        # each literal in a clause to get all of its neighbors.
        lit_to_clauses = defaultdict(set)
        for c in relevant_clauses:
            for lit in c:
                lit_to_clauses[abs(lit)].add(c)
        lit_to_clauses = dict(lit_to_clauses)

        def get_neighbors(clause):
            """ Return the set of clauses which have at least one variable in
            common with this one. """
            clause_sets = (lit_to_clauses[abs(lit)] for lit in clause)
            return sorted(set.union(*clause_sets), key=lambda c: c.lits)

        # If there aren't two end points then none of this makes any sense.
        # Just return what we have.
        if len(end_clauses) < 2:
            return (end_clauses,)

        ends = OrderedDict.fromkeys(end_clauses)
        ends = tuple(ends.keys())

        def jobs_in_path(path):
            return len([clause for clause in path
                        if clause.rule.reason.is_job])

        def tails(seq):
            for i in range(1, len(seq)):
                yield seq[i:]

        # Find a shortest path between each pair of end points if there is one
        raw_paths = list(itertools.chain.from_iterable(
            breadth_first_search(start, get_neighbors, rest)
            for start, rest in zip(ends, tails(ends))))

        # This is somewhat arbitrary, but for now we'll say that the best path
        # is the one with the most jobs in it.
        empty = (0, ())
        path_groups = itertools.groupby(
            sorted(raw_paths, key=jobs_in_path, reverse=True),
            key=jobs_in_path)
        equal_job_paths = tuple(next(path_groups, empty)[1])

        # Now start from the shortest and only include paths that include
        # something that we haven't seen before.
        seen_clauses = set()
        paths = []
        for path in sorted(equal_job_paths, key=len):
            p = set(path)
            if p - seen_clauses:
                seen_clauses.update(p)
                paths.append(path)

        return tuple(paths)

    def _end_clauses(self, relevant_clauses, implicand=None):
        """ Return the nodes which will serve as required points in our path.

        Given a bag of clauses, each possibly with rules and requirements
        attached, this pulls out the clauses whose rule came directly from a
        user request or whose variable contain the variable with the assignment
        conflict.
        """

        flat_clauses = set(
            c for c in relevant_clauses
            if c.rule and c.rule._requirements
            if implicand in c or c.rule.reason.is_job
        )
        roots = tuple(sorted(flat_clauses, key=self._key))
        return roots

    @property
    def rules(self):
        flat = (c.rule for path in self._conflict_paths for c in path)
        return tuple(OrderedDict.fromkeys(flat).keys())

    @property
    def requirements(self):
        # Every list of requirements start at a job, so only take the first one
        return tuple(OrderedDict.fromkeys(
            rule._requirements[0] for rule in self.rules))

    def clause_requirements(self, clause, ignore=None):
        """
        Return the user requirements that led to the creation of `clause`.

        If the clause hasn't been requested before, we search it and its
        parents recursively.
        """
        ignore = ignore or set()
        if clause in ignore:
            return []
        ignore.add(clause)
        if clause not in self._clause_requirements:
            # We haven't searched this clause before. Do so now.
            reqs = []
            if clause.rule and clause.rule._requirements:
                # This clause came directly from a rule that came from a
                # user requirement.
                reqs.append(clause)
            if clause.learned:
                # This clause is a learned synthesis of many other clauses. We
                # must follow them to find their requirements.
                trail = self.clause_trail(clause)
                reqs.extend(r for c in trail
                            for r in self.clause_requirements(c, ignore))
            # Memoize our result to avoid combinatorial explosion on recursive
            # calls
            self._clause_requirements[clause] = reqs
        return self._clause_requirements[clause]

    def clause_trail(self, clause, ignore=None):
        """
        Return the entire flattened list of clauses in this clause's trail.

        A learned clause has a "trail" of clauses which led to the learned
        clause being created. Clauses in this trail might also be learned
        clauses. This method recursively builds up all of non-learned clauses
        found by expanding and concatenating these trails.
        """
        ignore = ignore or set()
        if clause in ignore:
            return []
        ignore.add(clause)
        if clause not in self._flat_clause_trails:
            flat_trail = []
            if clause.learned:
                for t_clause in self._clause_trails[clause]:
                    if t_clause.learned:
                        flat_trail.extend(self.clause_trail(t_clause, ignore))
                    else:
                        flat_trail.append(t_clause)
                    ignore.add(t_clause)
            self._flat_clause_trails[clause] = flat_trail
        return self._flat_clause_trails[clause]

    def to_string(self, pool=None):
        # Build a string description of each conflict we've found
        return '\n'.join(
            self.string_from_clauses(clauses, pool=pool)
            for clauses in self._conflict_paths)

    def string_from_clauses(self, clauses, pool=None):
        details = OrderedDict()

        for clause in clauses:
            # Learned clauses have no meaningful explanation and thus, no Rule.
            # Instead, we grab the clauses from which it is derived.
            flat_clauses = self.clause_trail(clause) or (clause,)

            for clause in flat_clauses:
                # Add it to our set of clauses to include
                details.setdefault(self._key(clause), clause)

        reason = ["Conflicting requirements:"]
        for clause in details.values():
            if pool:
                reason.append(clause.rule.to_string(pool, unique=True))
            else:
                reason.append(str(clause.lits))

        return '\n'.join(reason) + '\n'


class MiniSATSolver(object):
    @classmethod
    def from_rules(cls, rules, policy=None):
        """
        Construct a SAT solver from a rules generator.

        Parameters
        ----------
        rules: RulesGenerator
        policy: IPolicy
            The policy to use for this SAT solver.

        Returns
        -------
        solver: MiniSATSolver.

        """
        solver = cls(policy)
        for rule in rules:
            solver.add_clause(rule.literals, rule=rule)
        solver._setup_assignments()
        return solver

    def __init__(self, policy=None):

        self.clauses = []
        self.watches = defaultdict(list)

        self.assignments = AssignmentSet()

        # The trail of clauses used to learn each new clause
        self.clause_trails = {}

        # A list of literals which become successively true (because of direct
        # assignment, or by unit propagation).
        self.levels = defaultdict(int)

        self.prop_queue = deque()

        # A list of all the decisions that we've made so far.
        self.trail = []

        # Keeps track of the independent assumptions so far. Each entry in
        # trail_lim is an index pointing to an assumption in trail. TODO:
        # Structure self.trail so that it keeps track of decisions per level.
        self.trail_lim = []

        # For each variable assignment, a reference to the clause that forced
        # this assignment.
        self.assigning_clauses = {}

        # Whether the system is satisfiable.
        self.status = None

        self._policy = policy or DefaultPolicy()

    def add_clause(self, clause, rule=None):
        """ Add a new clause to the solver.

        Parameters
        ----------
        clause : Clause
            The clause to add to the SAT problem
        rule : PackageRule
            An optional rule to associate with this clause. This is typically
            the rule from which the clause was derived.
        """
        # TODO: Do some simplifications, and check whether clause contains p
        # and -p at the same time.

        if not isinstance(clause, Clause):
            clause = Clause(clause, learned=False, rule=rule)

        if len(clause) == 0:
            # Clause is guaranteed to be false under the current variable
            # assignments.
            self.status = False
            return

        if len(clause) == 1:
            # Unit facts are enqueued.
            if not self.enqueue(clause[0], cause=clause):
                # Bail out if we've found a conflict
                conflict = UNSAT(
                    clause, clause,
                    self.clause_trails,
                    self.assigning_clauses)
                raise SatisfiabilityError(conflict)
        else:
            p, q = clause[:2]
            self.watches[-p].append(clause)
            self.watches[-q].append(clause)

        self.clauses.append(clause)

    def _setup_assignments(self):
        """Initialize assignments table.
        """
        variables = {abs(lit) for clause in self.clauses for lit in clause}
        assignments = self.assignments
        for variable in variables:
            if variable not in assignments:
                assignments[variable] = None

    def propagate(self):
        while len(self.prop_queue) > 0:
            lit = self.prop_queue.popleft()
            clauses = self.watches[lit]
            self.watches[lit] = []

            while len(clauses) > 0:
                clause = clauses.pop()
                unit = clause.rewatch(self.assignments, lit)

                # Re-insert in the appropriate watch list.
                self.watches[-clause.lits[1]].append(clause)

                # Deal with unit clauses.
                if unit is not None:
                    # TODO Refactor this to take into account the return value
                    # of enqueue().
                    if self.assignments.value(unit) is False:
                        # Conflict. Clear the queue and re-insert the remaining
                        # unwatched clauses into the watch list.
                        self.prop_queue.clear()
                        for remaining in clauses:
                            self.watches[lit].append(remaining)
                        return clause
                    else:
                        # Non-conflicting unit literal.
                        self.enqueue(unit, clause)

    def enqueue(self, lit, cause=None):
        """ Enqueue a new true literal. Return True if this assignment does not
        conflict with a previous assignment, otherwise False.

        Parameters
        ----------
        lit : literal (a signed integer)
            A literal to mark as True.
        cause : Clause
            An optional clause to associate with this assignment. This is
            typically the clause which forced the assignment via propagation.
        """
        status = self.assignments.value(lit)
        if status is not None:
            return status
        else:
            # New fact, store it.
            self.assignments[abs(lit)] = (lit > 0)

            self.prop_queue.append(lit)
            self.trail.append(lit)
            self.levels[abs(lit)] = self.decision_level
            self.assigning_clauses[abs(lit)] = cause
            return True

    def search(self):
        """ Return next solution or Raise SatisfiabilityError if unsatisfiable.
        """
        root_level = self.decision_level
        while True:
            conflict_clause = self.propagate()
            if conflict_clause is None:
                if self.number_assigned == self.number_variables:
                    # Model found.
                    return self.assignments.copy()  # Do something better...
                else:
                    # New variable decision.
                    p = self._policy.get_next_package_id(
                        self.assignments,
                        self.clauses,
                    )

                    self.assume(p)
            else:
                # Conflict!
                learned_clause, bt_level = self.analyze(conflict_clause)
                if root_level == self.decision_level:
                    conflict = UNSAT(
                        conflict_clause, learned_clause,
                        self.clause_trails,
                        self.assigning_clauses)
                    raise SatisfiabilityError(conflict)

                self.cancel_until(max(bt_level, root_level))
                self.record(learned_clause)

    def validate(self, solution_map):
        """Check whether a given set of assignments solves this SAT problem.
        """
        solution_literals = {variable if status else -variable
                             for variable, status in solution_map.items()}
        # True if any clause has no assigned literals and thus is undetermined
        has_unknown_clause = any(solution_literals.isdisjoint(clause.lits)
                                 for clause in self.clauses)
        return not has_unknown_clause

    def analyze(self, conflict):
        """ Produce a reason clause for a conflict.
        """
        p = None  # Will hold the UIP at the end of the search.

        # A tally of the number of literals encountered so far in the current
        # decision level, and downstream from the UIP.
        counter = 0
        # Variables that we've encountered during the search.
        seen = set()

        # Literals for the clause that we're learning.
        learned_lits = []
        # Level to backtrack to.
        btlevel = 0

        clause_trail = [conflict]

        while True:
            reason = conflict.calculate_reason(p)

            # Trace reason for current p.
            for lit in reason:
                var = abs(lit)
                if var not in seen:
                    seen.add(var)
                    if self.levels[var] == self.decision_level:
                        # A new literal on the current decision level.
                        counter += 1
                    else:
                        # At this point, we don't treat level 0 as
                        # special. Maybe that's a mistake...
                        learned_lits.append(-lit)
                        btlevel = max(btlevel, self.levels[var])

            # Select next literal to look at.
            while True:
                p = self.trail[-1]
                conflict = self.assigning_clauses[abs(p)]
                clause_trail.append(conflict)
                self.undo_one()
                if abs(p) in seen:
                    break

            counter -= 1
            if counter == 0:
                break

        learned_lits.append(-p)  # At this point p is the UIP.
        learned = Clause(learned_lits, learned=True)
        self.clause_trails[learned] = clause_trail
        return learned, btlevel

    def record(self, learned_clause):  # Needs test.
        """Drive the backtracking by adding a learned clause, which is unit by
        assumption.

        """
        # Reorder the learned clause, so that lits[0] is the asserting literal,
        # and lits[1] is the literal with highest decision level. This literal
        # will first become unbound by backtracking.
        lits = learned_clause.lits
        lits[0], lits[-1] = lits[-1], lits[0]

        # Index of the literal with the highest decision level.
        def key(arg):
            n, level = arg
            return level
        max_i = max(enumerate([self.levels.get(abs(lit), 0) for lit in lits]),
                    key=key)[0]
        if len(lits) >= 2:
            lits[1], lits[max_i] = lits[max_i], lits[1]

        self.add_clause(learned_clause)
        self.enqueue(learned_clause.lits[0], learned_clause)

    def undo_one(self):
        """Backtrack by one step.
        """
        p = self.trail.pop()
        v = abs(p)  # Underlying variable
        self.assignments[v] = None
        self.levels[v] = -1  # FIXME Why -1?

    def cancel_until(self, level):
        """Cancel all decisions up a given level.
        """
        while self.decision_level > level:
            self.cancel()

    def cancel(self):
        """Undo all decisions in the current decision level.
        """
        # Taken verbatim from Minisat paper.
        c = len(self.trail) - self.trail_lim.pop()
        for _ in range(c):
            self.undo_one()

    def assume(self, lit, cause=Clause([])):
        self.trail_lim.append(len(self.trail))  # FIXME: This is fishy.
        return self.enqueue(lit, cause=cause)

    @property
    def number_assigned(self):
        """ Return the number of currently assigned variables.
        """
        return self.assignments.num_assigned

    @property
    def number_variables(self):
        """ Return the number of variables in the SAT problem.
        """
        return len(self.assignments)

    @property
    def decision_level(self):
        """ Return the number of independent assumptions made so far.
        """
        return len(self.trail_lim)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import absolute_import

import contextlib
import shutil
import tempfile

from .timed_context import timed_context
from .graph import connected_nodes, toposort, transitive_neighbors
from ._collections import DefaultOrderedDict


@contextlib.contextmanager
def mkdtemp():
    d = tempfile.mkdtemp()
    yield d
    shutil.rmtree(d)



# -*- coding: utf-8 -*-

from collections import Counter

from .policy import IPolicy


class PolicyLogger(IPolicy):

    def __init__(self, policy, args=None, kwargs=None):
        self._policy = policy
        self._log_pool = args[0]
        self._log_installed = set(getattr(policy, '_installed_ids', ()))
        self._log_preferred = set(getattr(policy, '_preferred_ids', ()))
        self._log_args = args
        self._log_kwargs = kwargs
        self._log_required = []
        self._log_suggestions = []
        self._log_assignment_changes = []

    def get_next_package_id(self, assignments, clauses):
        self._log_assignment_changes.append(assignments.get_changelog())
        pkg_id = self._policy.get_next_package_id(assignments, clauses)
        self._log_suggestions.append(pkg_id)
        assignments.consume_changelog()
        return pkg_id

    def add_requirements(self, package_ids):
        self._log_required.extend(package_ids)
        self._log_installed.difference_update(package_ids)
        self._policy.add_requirements(package_ids)

    def _log_histogram(self, pkg_ids=None):
        if pkg_ids is None:
            pkg_ids = map(abs, self._log_suggestions)
        c = Counter(pkg_ids)
        lines = (
            "{:>25} {:>5}".format(self._log_pretty_pkg_id(k), v)
            for k, v in c.most_common()
        )
        pretty = '\n'.join(lines)
        return c, pretty

    def _log_pretty_pkg_id(self, pkg_id):
        package = self._log_pool.id_to_package(pkg_id)
        name_ver = '{} {}'.format(package.name, package.version)
        fill = '.' if pkg_id % 2 else ''
        try:
            repo = package.repository_info.name
        except AttributeError:
            repo = 'installed'
        return "{:{fill}<30} {:3} {}".format(name_ver, pkg_id, repo, fill=fill)

    def _log_report(self, with_assignments=True):

        def pkg_name(pkg_id):
            return pkg_key(pkg_id)[0]

        def pkg_key(pkg_id):
            pkg = self._log_pool.id_to_package(pkg_id)
            return pkg.name, pkg.version

        ids = map(abs, self._log_suggestions)
        report = []
        changes = []
        if self._log_assignment_changes and with_assignments:
            for pkg, change in self._log_assignment_changes[0].items():
                name = self._log_pretty_pkg_id(pkg)
                if change[1] is not None:
                    changes.append("{} : {}".format(name, change[1]))
            report.append('\n'.join(changes))

        required = set(self._log_required)
        installed = set(self._log_installed)
        for (i, sugg) in enumerate(ids):
            pretty = self._log_pretty_pkg_id(sugg)
            R = 'R' if sugg in required else ' '
            I = 'I' if sugg in installed else ' '
            change_str = ""
            try:
                items = self._log_assignment_changes[i + 1].items()
                sorted_items = sorted(items, key=lambda p: pkg_key(p[0]))
                if with_assignments:
                    changes = []
                    for pkg, change in sorted_items:
                        if pkg_name(pkg) != pkg_name(sugg):
                            _pretty = self._log_pretty_pkg_id(pkg)
                            fro, to = map(str, change)
                            msg = "{:10} - {:10} : {}"
                            changes.append(msg.format(fro, to, _pretty))
                    if changes:
                        change_str = '\n\t'.join([''] + changes)
                msg = "{:>4} {}{} - {}{}"
                report.append(msg.format(i, R, I, pretty, change_str))
                if any(v[1] is None for _, v in sorted_items):
                    report.append("BACKTRACKED\n")
            except IndexError:
                pass
        return '\n'.join(report)


def LoggedPolicy(policy_factory):
    def PolicyFactory(*args, **kwargs):
        policy = policy_factory(*args, **kwargs)
        logger = PolicyLogger(policy, args=args, kwargs=kwargs)
        return logger
    return PolicyFactory



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from ..assignment_set import AssignmentSet


class TestAssignmentSet(unittest.TestCase):

    def test_starts_empty(self):
        AS = AssignmentSet()
        self.assertEqual(AS.num_assigned, 0)
        self.assertEqual(len(AS), 0)
        self.assertEqual({}, AS.get_changelog())
        self.assertEqual([], AS.keys())
        self.assertEqual([], AS.values())
        self.assertEqual([], AS.items())

    def test_num_assigned(self):
        AS = AssignmentSet()

        AS[1] = None
        self.assertEqual(AS.num_assigned, 0)

        AS[2] = True
        self.assertEqual(AS.num_assigned, 1)

        AS[1] = False
        self.assertEqual(AS.num_assigned, 2)

        AS[2] = None
        self.assertEqual(AS.num_assigned, 1)

        AS[2] = True
        self.assertEqual(AS.num_assigned, 2)

        del AS[1]
        self.assertEqual(AS.num_assigned, 1)

        AS[2] = False
        self.assertEqual(AS.num_assigned, 1)

        AS[2] = None
        self.assertEqual(AS.num_assigned, 0)

        del AS[2]
        self.assertEqual(AS.num_assigned, 0)

    def test_container(self):
        AS = AssignmentSet()

        AS[1] = True
        AS[2] = False

        self.assertIn(1, AS)
        self.assertNotIn(3, AS)

        AS[4] = None
        AS[3] = True
        AS[5] = None

        self.assertIn(3, AS)
        self.assertIn(5, AS)

        del AS[5]
        self.assertNotIn(5, AS)

        expected = [(1, True), (2, False), (3, True), (4, None)]

        manual_result = list(zip(AS.keys(), AS.values()))
        self.assertEqual(AS.items(), expected)
        self.assertEqual(manual_result, expected)
        self.assertEqual(len(AS), len(expected))

    def test_copy(self):

        AS = AssignmentSet()

        AS[1] = None
        AS[2] = True
        AS[3] = None
        AS[4] = False
        AS[5] = True

        expected = {
            1: None,
            2: True,
            3: None,
            4: False,
            5: True,
        }

        copied = AS.copy()

        self.assertIsNot(copied._data, AS._data)
        self.assertEqual(copied.to_dict(), expected)

        expected = {k: None for k, v in expected.items() if v is not None}

        self.assertIsNot(copied._orig, AS._orig)
        self.assertEqual(copied._orig, expected)

        self.assertEqual(copied.num_assigned, AS.num_assigned)

        del AS[2]
        self.assertIn(2, copied)

    def test_value(self):
        AS = AssignmentSet()

        AS[1] = False
        AS[2] = True
        AS[3] = None

        self.assertTrue(AS.value(-1))
        self.assertTrue(AS.value(2))

        self.assertFalse(AS.value(1))
        self.assertFalse(AS.value(-2))

        self.assertIs(AS.value(3), None)
        self.assertIs(AS.value(-3), None)

        del AS[2]
        self.assertIs(AS.value(-2), None)
        self.assertIs(AS.value(2), None)

        AS[1] = None
        self.assertIs(AS.value(-1), None)
        self.assertIs(AS.value(1), None)

        AS[3] = True
        self.assertIs(AS.value(-3), False)
        self.assertIs(AS.value(3), True)

        AS[3] = False
        self.assertIs(AS.value(-3), True)
        self.assertIs(AS.value(3), False)

    def test_getitem(self):
        AS = AssignmentSet()

        AS[1] = False
        AS[2] = True
        AS[3] = None

        self.assertFalse(AS[1])
        self.assertTrue(AS[2])
        self.assertIs(AS[3], None)

        with self.assertRaises(KeyError):
            AS[4]

        self.assertFalse(AS.get(1))
        self.assertTrue(AS.get(2))
        self.assertIs(AS.get(3), None)
        self.assertIs(AS.get(4), None)

    def test_changelog(self):

        AS = AssignmentSet()

        AS[1] = None

        expected = {}
        self.assertEqual(AS.get_changelog(), expected)

        AS[2] = True
        expected[2] = (None, True)
        self.assertEqual(AS.get_changelog(), expected)

        AS[2] = False
        expected[2] = (None, False)
        self.assertEqual(AS.get_changelog(), expected)

        del AS[2]
        del expected[2]
        self.assertEqual(AS.get_changelog(), expected)

        # Keep should preserve the log
        self.assertEqual(AS.get_changelog(), expected)
        self.assertEqual(AS.get_changelog(), expected)

        # Otherwise clear the log
        log = AS.consume_changelog()
        self.assertEqual(log, expected)
        self.assertEqual(AS.get_changelog(), {})

        # Test when something is already present
        AS[1] = False
        expected = {1: (None, False)}
        self.assertEqual(AS.get_changelog(), expected)

        del AS[1]
        expected = {}
        self.assertEqual(AS.get_changelog(), expected)



import abc
import six
import enum

from okonomiyaki.versions import EnpkgVersion


class ConstraintKinds(enum.Enum):
    install_requires = 'install_requires'
    conflicts = 'conflicts'
    provides = 'provides'


class IRepositoryInfo(six.with_metaclass(abc.ABCMeta)):
    @abc.abstractproperty
    def name(self):
        """ A name that uniquely indentifies a repository."""


class RepositoryInfo(IRepositoryInfo):
    def __init__(self, name):
        self._name = name
        self._key = (name,)
        self._hash = hash(self._key)

    @property
    def name(self):
        """A name that uniquely indentifies a repository."""
        return self._name

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return self._key == other._key

    def __ne__(self, other):
        return self._key != other._key

    def __repr__(self):
        return "Repository(<{0.name}>)".format(self)


class PackageMetadata(object):
    """ PackageMetadata represents an immutable, versioned Python distribution
    and its relationship with other packages.
    """

    @classmethod
    def _from_pretty_string(cls, s):
        """ Create an instance from a pretty string.

        A pretty string looks as follows::

            'numpy 1.8.1-1; depends (MKL ^= 10.3)'

        Note
        ----
        Don't use this in production code, only meant to be used for testing.
        """
        # FIXME: local import to workaround circular imports
        from .constraints import PrettyPackageStringParser
        parser = PrettyPackageStringParser(EnpkgVersion.from_string)
        return parser.parse_to_package(s)

    def __init__(self, name, version, install_requires=None, conflicts=None,
                 provides=None):
        """ Return a new PackageMetdata object.

        Parameters
        ----------
        name : str
            The name of the Python distribution, e.g. "numpy"
        version : EnpkgVersion
            An EnpkgVersion object describing the version of this package.
        install_requires : tuple(tuple(str, tuple(tuple(str))))
            A tuple of tuples mapping distribution names to disjunctions of
            conjunctions of version constraints.

            For example, consider a package that depends on the following:
                - nose
                - six (> 1.2, <= 1.2.3), or >= 1.2.5-2
                    Written as intervals, (1.2, 1.2.3] or [1.2.5-2, \infty)
                - MKL >= 10.1, < 11

            The constraint tuple representing this would be:
                (("MKL", ((">= 10.1", "< 11"),)),
                 ("nose", (("*",),)),
                 ("six", (("> 1.2", "<= 1.2.3"), (">= 1.2.5-2",)))
        conflicts : tuple(tuple(str, tuple(tuple(str))))
            A tuple of tuples mapping distribution names to disjunctions of
            conjunctions of version constraints.

            This works the same way as install_requires, but instead denotes
            packages that must *not* be installed with this package.
        provides :  tuple(tuple(str, tuple(tuple(str))))
            A tuple of tuples mapping package and virtual package names to
            disjunctions of conjunctions of version constraints.

            For example, consider a package ``numpy-nomkl`` which should be a
            drop-in replacement for the normal ``numpy`` or ``numeric``
            packages. ``numpy-nomkl`` would have the following `provides`:
                (("numpy", (("*",),)), ("numeric", (("*",),)))

            At this time, no version constraint is permitted for names
            specified in `provides`.
        """
        self._name = name
        self._provides = tuple(provides or ())
        self._version = version
        self._install_requires = install_requires or ()
        self._conflicts = conflicts or ()
        self._key = (name, version, self._install_requires, self._conflicts)
        self._hash = hash(self._key)

    @property
    def name(self):
        return self._name

    @property
    def provides(self):
        constraint_str = "*"
        this_pkg = ((self._name, ((constraint_str,),)),)
        return this_pkg + self._provides

    @property
    def version(self):
        return self._version

    @property
    def install_requires(self):
        return self._install_requires

    @property
    def conflicts(self):
        return self._conflicts

    def __repr__(self):
        return "{0}('{1}-{2}')".format(
            self.__class__.__name__, self._name, self._version)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        try:
            return self._key == other._key
        except AttributeError:
            return NotImplemented

    def __ne__(self, other):
        try:
            return self._key != other._key
        except AttributeError:
            return NotImplemented


class RepositoryPackageMetadata(object):
    @classmethod
    def _from_pretty_string(cls, s, repository_info):
        package = PackageMetadata._from_pretty_string(s)
        return cls(package, repository_info)

    def __init__(self, package, repository_info):
        self._package = package
        self._repository_info = repository_info

        self._key = (package._key, repository_info)
        self._hash = hash(self._key)

    @property
    def name(self):
        return self._package.name

    @property
    def provides(self):
        return self._package.provides

    @property
    def version(self):
        return self._package.version

    @property
    def install_requires(self):
        return self._package.install_requires

    @property
    def conflicts(self):
        return self._package.conflicts

    @property
    def repository_info(self):
        return self._repository_info

    def __repr__(self):
        return (
            "RepositoryPackageMetadata('{pkg._name}-{pkg._version}'"
            ", repo={repository_info!r})".format(
                pkg=self._package, repository_info=self._repository_info))

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        try:
            return self._key == other._key
        except AttributeError:
            return NotImplemented

    def __ne__(self, other):
        try:
            return self._key != other._key
        except AttributeError:
            return NotImplemented



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import six

from .policy import IPolicy, pkg_id_to_version
from .policy_logger import LoggedPolicy


class UndeterminedClausePolicy(IPolicy):

    """ An IPolicy that gathers all undetermined packages from clauses whose
    truth value is not yet known and suggests them in descending order by
    package version number. """

    def __init__(self, pool, installed_repository,
                 ignore_installed_packages=None):
        if ignore_installed_packages is None:
            ignore_installed_packages = set()
        self._pool = pool

        by_version = six.functools.partial(pkg_id_to_version, self._pool)
        installed_packages = set(installed_repository)
        prefer_installed_pkgs = installed_packages - ignore_installed_packages
        self._prefer_installed_pkg_ids = sorted(
            (pool.package_id(pkg) for pkg in prefer_installed_pkgs),
            key=by_version)

        self._decision_set = set()
        self._requirements = set()
        self._all_ids = set()

    def add_requirements(self, package_ids):
        self._requirements.update(package_ids)

    def get_next_package_id(self, assignments, clauses):
        """Get the next unassigned package.
        """
        if assignments.new_keys:
            self._refresh_decision_set(assignments, clauses)

        candidate_id = None
        best = self._best_candidate

        candidate_id = self._best_sorted_candidate(
            self._prefer_installed_pkg_ids, assignments)

        if candidate_id is None:
            candidate_id = best(self._requirements, assignments)

        if candidate_id is None:
            candidate_id = best(self._decision_set, assignments, update=True)

        if candidate_id is None:
            self._refresh_decision_set(assignments, clauses)
            candidate_id = best(self._decision_set, assignments)

        if candidate_id is None:
            candidate_id = best(self._all_ids, assignments)

        assert assignments.get(candidate_id) is None, \
            "Trying to assign to a variable which is already assigned."

        return candidate_id

    def _without_assigned(self, package_ids, assignments):
        return package_ids.difference(assignments.assigned_ids)

    def _best_sorted_candidate(self, package_ids, assignments):
        for p_id in package_ids:
            if p_id not in assignments.assigned_ids:
                return p_id

    def _best_candidate(self, package_ids, assignments, update=False):
        by_version = six.functools.partial(pkg_id_to_version, self._pool)
        unassigned = self._without_assigned(package_ids, assignments)
        if update:
            package_ids.clear()
            package_ids.update(unassigned)
        try:
            return max(unassigned, key=by_version)
        except ValueError:
            return None

    def _refresh_decision_set(self, assignments, clauses):
        assignments.consume_changelog()

        all_ids = {abs(l) for c in clauses for l in c.lits}
        all_ids.update(self._prefer_installed_pkg_ids)
        self._all_ids = all_ids

        unsatisfied_clauses = {
            clause for clause in clauses
            if not any(assignments.value(l) for l in clause.lits)
        }
        self._decision_set.clear()
        self._decision_set.update(
            abs(lit)
            for clause in unsatisfied_clauses
            for lit in clause.lits
        )
        self._decision_set.difference_update(assignments.assigned_ids)


LoggedUndeterminedClausePolicy = LoggedPolicy(UndeterminedClausePolicy)



import os.path
import sys
import textwrap

import attr
import six

from ..constraints import ConflictRequirement, InstallRequirement
from ..package import RepositoryInfo, RepositoryPackageMetadata
from ..request import _Job, JobType
from ..test_utils import Scenario, parse_package_list, repository_factory
from ..utils import mkdtemp

if sys.version_info[0] == 2:
    import unittest2 as unittest
else:
    import unittest


P = RepositoryPackageMetadata._from_pretty_string
CR = ConflictRequirement._from_string
IR = InstallRequirement._from_string


class TestRepositoryFactory(unittest.TestCase):
    def test_simple(self):
        # Given
        package_strings = [
            u"MKL 10.3-1",
            u"numpy 1.8.1-1; depends (MKL ^= 10.3)",
            u"numpy 1.8.1-2; depends (MKL ^= 10.3)",
        ]
        repository_packages = [
            "MKL 10.3-1",
            "numpy 1.8.1-2",
        ]
        repository_info = RepositoryInfo(u"acme/loony")
        r_numpy = P(
            u"numpy 1.8.1-2; depends (MKL ^= 10.3)", repository_info,
        )

        # When
        packages = dict(parse_package_list(package_strings))
        repository = repository_factory(repository_packages, repository_info,
                                        packages)

        # Then
        self.assertEqual(len(repository), 2)
        self.assertEqual(len(repository.find_packages("numpy")), 1)

        numpy = repository.find_packages("numpy")[0]
        self.assertEqual(numpy, r_numpy)


class TestScenario(unittest.TestCase):
    def test_simple(self):
        # Given
        yaml = six.StringIO(textwrap.dedent("""\
        packages:
            - MKL 10.3-1
            - numpy 1.8.1-1; depends (MKL ^= 10.3)
            - numpy 1.8.1-2; depends (MKL ^= 10.3)

        remote:
            - MKL 10.3-1
            - numpy 1.8.1-2

        request:
            - operation: install
              requirement: numpy
        """))
        r_jobs = [_Job(IR("numpy"), JobType.install)]

        # When
        scenario = Scenario.from_yaml(yaml)

        # Then
        self.assertEqual(len(scenario.remote_repositories), 1)
        remote_repository = scenario.remote_repositories[0]
        self.assertEqual(len(remote_repository), 2)

        self.assertEqual(len(scenario.installed_repository), 0)

        jobs = scenario.request.jobs
        self.assertEqual(jobs, r_jobs)

    def test_from_filename(self):
        # Given
        data = textwrap.dedent("""\
        packages:
            - MKL 10.3-1
            - numpy 1.8.1-1; depends (MKL ^= 10.3)
            - numpy 1.8.1-2; depends (MKL ^= 10.3)

        remote:
            - MKL 10.3-1
            - numpy 1.8.1-2

        request:
            - operation: install
              requirement: numpy
        """)
        r_jobs = [_Job(IR("numpy"), JobType.install)]

        # When
        with mkdtemp() as d:
            path = os.path.join(d, "scenario.yaml")
            with open(path, "wt") as fp:
                fp.write(data)
            scenario = Scenario.from_yaml(path)

        # Then
        self.assertEqual(len(scenario.remote_repositories), 1)
        remote_repository = scenario.remote_repositories[0]
        self.assertEqual(len(remote_repository), 2)

        self.assertEqual(len(scenario.installed_repository), 0)

        jobs = scenario.request.jobs
        self.assertEqual(jobs, r_jobs)

    def test_simple_marked(self):
        # Given
        yaml = six.StringIO(textwrap.dedent("""\
        packages:
            - MKL 10.3-1

        remote:
            - MKL 10.3-1

        installed:
            - MKL 10.3-1

        marked:
            - MKL
        """))
        r_jobs = [_Job(IR("MKL"), JobType.install)]

        # When
        scenario = Scenario.from_yaml(yaml)

        # Then
        jobs = scenario.request.jobs
        self.assertEqual(jobs, r_jobs)

    def test_modify_marked(self):
        # Given
        yaml = six.StringIO(textwrap.dedent("""\
        packages:
            - MKL 10.3-1

        remote:
            - MKL 10.3-1

        installed:
            - MKL 10.3-1

        marked:
            - MKL

        request:
            - operation: remove
              requirement: MKL
        """))
        r_jobs = [_Job(CR("MKL"), JobType.remove)]

        # When
        scenario = Scenario.from_yaml(yaml)

        # Then
        jobs = scenario.request.jobs
        self.assertEqual(jobs, r_jobs)

    def test_load_modifiers(self):
        # Given
        yaml = six.StringIO(textwrap.dedent("""\
        packages:
            - MKL 10.3-1

        modifiers:
            allow_newer: [MKL]
            allow_older:
                - numpy
            allow_any:
                - pyzmq
                - pandas

        request:
            - operation: install
              requirement: numpy
        """))
        expected = {
            'allow_newer': set(['MKL']),
            'allow_older': set(['numpy']),
            'allow_any': set(['pyzmq', 'pandas']),
        }

        # When
        scenario = Scenario.from_yaml(yaml)

        # Then
        constraints = attr.asdict(scenario.request.modifiers, recurse=False)
        self.assertEqual(constraints, expected)



from __future__ import absolute_import

from collections import OrderedDict


class Constraint(object):
    pass


class Clause(Constraint):

    def __init__(self, lits, learned=False, rule=None):
        """
        Create a new Clause.

        Parameters
        ----------
        lits : list of literals
            The literals in this clause.
        learned : bool
            A flag indicated whether this clause was learned during solving.
        rule : PackageRule
            A package rule associated with the clause. Typically the rule from
            which this clause was created.
        """
        self.learned = learned
        self.rule = rule
        # This maintains the ordering while removing duplicate values
        self.lits = list(OrderedDict.fromkeys(lits).keys())

    def rewatch(self, assignments, lit):
        """Find a new literal to watch.

        The running assumption is that watched literals should either be True
        or not assigned (None). If a watched literal becomes False, a new watch
        must be found. When this is not possible, the other watched literal
        is necessarily True (unit information).

        Returns
        -------
        unit_information: Literal or None
            A literal which has become true under the current assignments.

        Note that this method does not check whether unit propagation leads to
        a conflict.

        """
        # Internal assumption: the watched literals are the first elements of
        # lits. If necessary, this method will re-order some of the literals to
        # keep this assumption.
        lits = self.lits
        assert -lit in lits[:2]

        if lits[0] == -lit:
            lits[0], lits[1] = lits[1], -lit

        if assignments.value(lits[0]) is True:
            # This clause has been satisfied, add it back to the watch list. No
            # unit information can be deduced.
            return None

        # Look for another literal to watch, and switch it with lit[1] to keep
        # the assumption on the watched literals in place.
        for n, other in enumerate(lits[2:]):
            if assignments.value(other) is not False:
                # Found a new literal that could serve as a watch.
                lits[1], lits[n + 2] = other, -lit
                return None

        # Clause is unit under assignment. Return the literal that can be
        # propagated.
        return lits[0]

    def calculate_reason(self, p=None):
        """For a conflicting clause, return the reason for propagating p.

        For example, if the clause is x \/ y \/ z, then the reason for
        propagating x is -y /\ -z. By convention, f the literal p does not
        occur in the clause, the negative of the whole clause is returned.

        """
        # TODO: We can speed this up if we can guarantee that we'll only ask
        # for the reason of the first literal, as is the case in the Minisat
        # paper.
        return [-lit for lit in self.lits if lit != p]

    def __len__(self):
        return len(self.lits)

    def __getitem__(self, s):
        return self.lits[s]

    def __repr__(self):
        return "Clause({}, learned={})".format(self.lits, self.learned)

    def __lt__(self, other):
        raise TypeError("no ordering relation is defined for clauses")



import unittest

import re
import six

from okonomiyaki.versions import EnpkgVersion

from simplesat.constraints import PrettyPackageStringParser, InstallRequirement
from simplesat.errors import InvalidConstraint
from simplesat.repository import Repository
from simplesat.request import Request

from ..pool import Pool


V = EnpkgVersion.from_string


NUMPY_PACKAGES = u"""\
mkl 10.2-1
mkl 10.2-2
mkl 10.3-1
numpy 1.4.0-1
numpy 1.4.0-2
numpy 1.4.0-4; depends (MKL ^= 10.2)
numpy 1.4.0-6; depends (MKL == 10.2-2)
numpy 1.4.0-7; depends (MKL == 10.2-2)
numpy 1.4.0-8; depends (MKL == 10.2-2)
numpy 1.4.0-9; depends (MKL == 10.2-2)
numpy 1.5.1-1; depends (MKL == 10.3-1)
numpy 1.5.1-2; depends (MKL == 10.3-1)
numpy 1.6.0-0
numpy 1.6.0-1; depends (MKL == 10.3-1)
numpy 1.6.0-2; depends (MKL == 10.3-1)
numpy 1.6.0-3; depends (MKL == 10.3-1)
numpy 1.6.0-4; depends (MKL == 10.3-1)
numpy 1.6.0-5; depends (MKL == 10.3-1)
numpy 1.6.0b2-1; depends (MKL == 10.3-1)
numpy 1.6.1-1; depends (MKL == 10.3-1)
numpy 1.6.1-2; depends (MKL == 10.3-1)
numpy 1.6.1-3; depends (MKL == 10.3-1)
numpy 1.6.1-5; depends (MKL == 10.3-1)
numpy 1.7.1-1; depends (MKL == 10.3-1)
numpy 1.7.1-2; depends (MKL == 10.3-1)
numpy 1.7.1-3; depends (MKL == 10.3-1)
numpy 1.8.0-1; depends (MKL == 10.3-1)
numpy 1.8.0-2; depends (MKL == 10.3-1)
numpy 1.8.0-3; depends (MKL == 10.3-1)
numpy 1.8.1-1; depends (MKL == 10.3-1)
"""


class TestPool(unittest.TestCase):
    def packages_from_definition(self, packages_definition):
        parser = PrettyPackageStringParser(EnpkgVersion.from_string)

        return [
            parser.parse_to_package(line)
            for line in packages_definition.splitlines()
        ]

    def test_what_provides_caret(self):
        # Given
        repository = Repository(self.packages_from_definition(NUMPY_PACKAGES))
        requirement = InstallRequirement._from_string("numpy ^= 1.8.1")

        # When
        pool = Pool([repository])
        candidates = pool.what_provides(requirement)

        # Then
        self.assertEqual(len(candidates), 1)
        self.assertEqual(candidates[0].version, V("1.8.1-1"))

    def test_what_provides_casing(self):
        # Given
        repository = Repository(self.packages_from_definition(NUMPY_PACKAGES))
        requirement = InstallRequirement._from_string("mkl ^= 10.2")

        # When
        pool = Pool([repository])
        candidates = pool.what_provides(requirement)
        versions = [str(candidate.version) for candidate in candidates]

        # Then
        six.assertCountEqual(self, versions, ["10.2-1", "10.2-2"])

    def test_what_provides_simple(self):
        # Given
        repository = Repository(self.packages_from_definition(NUMPY_PACKAGES))
        requirement = InstallRequirement._from_string("numpy >= 1.8.0")

        # When
        pool = Pool([repository])
        candidates = pool.what_provides(requirement)
        versions = [str(candidate.version) for candidate in candidates]

        # Then
        six.assertCountEqual(
            self, versions, ["1.8.0-1", "1.8.0-2", "1.8.0-3", "1.8.1-1"]
        )

    def test_what_provides_multiple(self):
        # Given
        repository = Repository(self.packages_from_definition(NUMPY_PACKAGES))
        requirement = InstallRequirement._from_string(
            "numpy >= 1.8.0, numpy < 1.8.1")

        # When
        pool = Pool([repository])
        candidates = pool.what_provides(requirement)
        versions = [str(candidate.version) for candidate in candidates]

        # Then
        six.assertCountEqual(
            self, versions, ["1.8.0-1", "1.8.0-2", "1.8.0-3"]
        )

    def test_id_to_string(self):
        # Given
        repository = Repository(self.packages_from_definition(NUMPY_PACKAGES))
        requirement = InstallRequirement._from_string("numpy >= 1.8.1")

        # When
        pool = Pool([repository])
        candidate = pool.what_provides(requirement)[0]
        package_id = pool.package_id(candidate)

        # Then
        self.assertEqual(pool.id_to_string(package_id), "+numpy-1.8.1-1")
        self.assertEqual(pool.id_to_string(-package_id), "-numpy-1.8.1-1")

    def test_modification_what_provides(self):
        # Given
        repository = Repository(self.packages_from_definition(
            "numpy 1.8.1-1; depends (MKL == 10.3-1)"))
        request = Request()
        request.modifiers.allow_newer.add('numpy')

        # When
        pool = Pool([repository], modifiers=request.modifiers)
        requirement = InstallRequirement._from_string('numpy ^= 1.7')
        numpy_181 = list(repository)[0]
        result = pool.what_provides(requirement)
        expected = [numpy_181]

        # Then
        self.assertEqual(result, expected)

    def test_reject_version_constraint_on_provides_metadata(self):

        # Given
        constraints = (
            u"A > 1.8",
            u"A >= 1.8",
            u"A <= 1.8",
            u"A < 1.8",
            u"A ^= 1.8",
            u"A == 1.8-1",
        )

        # When
        for constraint in constraints:
            repository = Repository(self.packages_from_definition(
                u"numpy 1.9.2-1; provides ({})".format(constraint)))
            constraint_re = re.escape(constraint)

            # Then
            with self.assertRaisesRegexp(InvalidConstraint, constraint_re):
                Pool([repository])

    def test_accept_anyversion_constraint_on_provides_metadata(self):
        # When
        repository = Repository(self.packages_from_definition(
            u"numpy 1.9.2-1; provides (A, B *)"))

        # Then
        Pool([repository])

    def test_iter_packages(self):
        # Given
        numpy_packages = self.packages_from_definition(NUMPY_PACKAGES)
        repository = Repository(numpy_packages)

        # When
        pool = Pool([repository])

        # Then
        packages = set(pool.iter_packages())
        self.assertEqual(packages, set(numpy_packages))

    def test_iter_package_ids(self):
        # Given
        numpy_packages = self.packages_from_definition(NUMPY_PACKAGES)
        repository = Repository(numpy_packages)

        # When
        pool = Pool([repository])

        # Then
        package_ids = set(pool.iter_package_ids())
        self.assertEqual(package_ids, set(pool._id_to_package_.keys()))



import sys
import unittest

from okonomiyaki.versions import EnpkgVersion

from simplesat.constraints.kinds import (
    Any, EnpkgUpstreamMatch, Equal, GT, GEQ, LT, LEQ, Not
)
from simplesat.constraints.parser import (
    _RawConstraintsParser, _RawRequirementParser
)
from simplesat.errors import SolverException


V = EnpkgVersion.from_string


class Test_RawConstraintsParser(unittest.TestCase):
    def setUp(self):
        self.parser = _RawConstraintsParser()

    def _parse(self, s):
        return self.parser.parse(s, V)

    def test_empty(self):
        # Given
        constraints_string = ""

        # When
        constraints = self._parse(constraints_string)

        # Then
        self.assertEqual(constraints, (Any(),))

    def test_simple(self):
        # Given
        constraints_string = "> 1.2.0-1"
        r_constraints = (GT(V("1.2.0-1")),)

        # When
        constraints = self._parse(constraints_string)

        # Then
        self.assertEqual(constraints, r_constraints)

        # Given
        constraints_string = ">= 1.2.0-1"
        r_constraints = (GEQ(V("1.2.0-1")),)

        # When
        constraints = self._parse(constraints_string)

        # Then
        self.assertEqual(constraints, r_constraints)

        # Given
        constraints_string = "<= 1.2.0-1"
        r_constraints = (LEQ(V("1.2.0-1")),)

        # When
        constraints = self._parse(constraints_string)

        # Then
        self.assertEqual(constraints, r_constraints)

        # Given
        constraints_string = "< 1.2.0-1"
        r_constraints = (LT(V("1.2.0-1")),)

        # When
        constraints = self._parse(constraints_string)

        # Then
        self.assertEqual(constraints, r_constraints)

        # Given
        constraints_string = "^= 1.2.0-1"
        r_constraints = (EnpkgUpstreamMatch(V("1.2.0-1")),)

        # When
        constraints = self._parse(constraints_string)

        # Then
        self.assertEqual(constraints, r_constraints)

    def test_multiple(self):
        # Given
        constraints_string = ">= 1.2.0-1, < 1.4, != 1.3.8-1"
        r_constraints = (
            GEQ(V("1.2.0-1")), LT(V("1.4")), Not(V("1.3.8-1"))
        )

        # When
        constraints = self._parse(constraints_string)

        # Then
        self.assertEqual(constraints, r_constraints)

    def test_invalid_string(self):
        # Given
        constraints_string = ">= 1.2.0-1 123"

        # When/Then
        with self.assertRaises(SolverException):
            self._parse(constraints_string)


class Test_RawRequirementParser(unittest.TestCase):
    def setUp(self):
        self.parser = _RawRequirementParser()

    def _parse(self, s):
        return self.parser.parse(s, V)

    def test_simple(self):
        # Given
        requirement_string = "numpy == 1.8.1-1"
        r_constraints = {"numpy": tuple([Equal(V("1.8.1-1"))])}

        # When
        constraints = self._parse(requirement_string)

        # Then
        self.assertEqual(constraints, r_constraints)

    def test_multiple(self):
        # Given
        requirement_string = "numpy >= 1.8.1, numpy < 1.9.0"
        r_constraints = {"numpy": (GEQ(V("1.8.1-0")), LT(V("1.9.0")))}

        # When
        constraints = self._parse(requirement_string)

        # Then
        self.assertEqual(constraints, r_constraints)

    def test_multiple_names(self):
        # Given
        requirement_string = "numpy >= 1.8.1, scipy >= 0.14.0"
        r_constraints = {"numpy": (GEQ(V("1.8.1-0")),),
                         "scipy": (GEQ(V("0.14.0")),)}

        # When
        constraints = self._parse(requirement_string)

        # Then
        self.assertEqual(constraints, r_constraints)

    def test_no_version(self):
        # Given
        requirement_string = "numpy"
        r_constraints = {"numpy": tuple([Any()])}

        # When
        constraints = self._parse(requirement_string)

        # Then
        self.assertEqual(constraints, r_constraints)

        # Given
        requirement_string = "MKL == 10.3-1, numpy"
        r_constraints = {"numpy": tuple([Any()]),
                         "MKL": tuple([Equal(V("10.3-1"))])}

        # When
        constraints = self._parse(requirement_string)

        # Then
        self.assertEqual(constraints, r_constraints)

        # Given
        requirement_string = "scikits.statsmodels"
        r_constraints = {"scikits.statsmodels": tuple([Any()])}

        # When
        constraints = self._parse(requirement_string)

        # Then
        self.assertEqual(constraints, r_constraints)

        # Given
        requirement_string = "special_package.123"
        r_constraints = {"special_package.123": tuple([Any()])}

        # When
        constraints = self._parse(requirement_string)

        # Then
        self.assertEqual(constraints, r_constraints)

    def test_invalid(self):
        # Given
        requirement_string = "numpy >= "

        # When/Then
        with self.assertRaises(SolverException):
            self._parse(requirement_string)

        # Given
        requirement_string = "numpy-no-mkl"

        # When/Then
        with self.assertRaises(SolverException):
            self._parse(requirement_string)

        # Given
        requirement_string = "numpy mkl"

        # When/Then
        with self.assertRaises(SolverException):
            self._parse(requirement_string)



# -*- coding: utf-8 -*-

import io
import mock
import unittest

from simplesat.errors import MissingConflicts, MissingInstallRequires

from ..pool import Pool
from ..rules_generator import RuleType, RulesGenerator
from ..test_utils import Scenario


class TestRulesGenerator(unittest.TestCase):

    def test_prefer_installed(self):

        # Given
        yaml = u"""
            packages:
                - A 1.0.0-1
            installed:
                - A 1.0.0-1
            marked:
                - A
            request:
                - operation: "update_all"
        """
        scenario = Scenario.from_yaml(io.StringIO(yaml))

        # When
        repos = list(scenario.remote_repositories)
        repos.append(scenario.installed_repository)
        pool = Pool(repos)
        installed_package_ids = {
            pool.package_id(p): p for p in scenario.installed_repository}
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids)
        rules = list(rules_generator.iter_rules())

        # Then
        self.assertEqual(len(rules), 2)

        # Given/When
        update = rules[1]

        # Then
        self.assertEqual(update.reason, RuleType.job_update)
        self.assertEqual(len(update.literals), 1)

        # Given
        installed_repo_package = next(iter(scenario.installed_repository))

        # When
        pkg_id = update.literals[0]
        package = pool.id_to_package(pkg_id)

        # Then
        self.assertIs(package, installed_repo_package)

    def test_provides(self):
        yaml = u"""
            packages:
              - A 1.0-1; provides (X)
              - B 1.0-1; provides (X)
              - C 1.0-1; provides (X)
              - X 1.0-1
              - Z 1.0; depends (X)

            request:
              - operation: "install"
                requirement: "Z"
        """

        scenario = Scenario.from_yaml(io.StringIO(yaml))

        # When
        repos = list(scenario.remote_repositories)
        repos.append(scenario.installed_repository)
        pool = Pool(repos)
        installed_package_ids = {
            pool.package_id(p): p for p in scenario.installed_repository}
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids)
        rules = list(rules_generator.iter_rules())

        # Then
        self.assertEqual(len(rules), 2)

        # Given/When
        rule = rules[0]
        r_literals = (-5, 1, 2, 3, 4)

        # Then
        self.assertEqual(rule.reason, RuleType.package_requires)
        self.assertEqual(rule.literals, r_literals)

    def test_conflicts(self):
        # Given
        yaml = u"""
            packages:
              - quark 1.0.1-2
              - atom 1.0.0-1; depends (quark > 1.0); conflicts (gdata ^= 1.0.0)
              - atom 1.0.1-1; depends (quark)
              - gdata 1.0.0-1; conflicts (atom >= 1.0.1)

            request:
              - operation: "install"
                requirement: "atom"
              - operation: "install"
                requirement: "gdata"
        """
        scenario = Scenario.from_yaml(io.StringIO(yaml))

        # When
        repos = list(scenario.remote_repositories)
        repos.append(scenario.installed_repository)
        pool = Pool(repos)
        installed_package_ids = {
            pool.package_id(p): p for p in scenario.installed_repository}
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids)
        rules = list(rules_generator.iter_rules())

        # Then
        self.assertEqual(len(rules), 7)

        # Given/When
        conflict = rules[2]
        r_literals = (-3, -1)

        # Then
        self.assertEqual(conflict.reason, RuleType.package_conflicts)
        self.assertEqual(conflict.literals, r_literals)

        # Given/When
        conflict = rules[5]
        r_literals = (-3, -2)

        # Then
        self.assertEqual(conflict.reason, RuleType.package_conflicts)
        self.assertEqual(conflict.literals, r_literals)

    def test_missing_direct_dependencies_package(self):
        # Given
        yaml = u"""
            packages:
              - atom 1.0.0-1; depends (quark > 1.0); conflicts (gdata ^= 1.0.0)
              - gdata 1.0.0-1; conflicts (atom >= 1.0.1)

            request:
              - operation: "install"
                requirement: "atom"
        """
        scenario = Scenario.from_yaml(io.StringIO(yaml))
        expected_log = ("Blocking package 'atom 1.0.0-1'" " from 'remote': no"
                        " candidates found for dependency 'quark > 1.0-0'")
        expected_rule = (-1,)

        # When
        repos = list(scenario.remote_repositories)
        repos.append(scenario.installed_repository)
        pool = Pool(repos)
        installed_package_ids = {
            pool.package_id(p): p for p in scenario.installed_repository}
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids)
        with mock.patch('simplesat.rules_generator.logger') as mock_logger:
            rules = list(rules_generator.iter_rules())

        # Then
        result_log = mock_logger.info.call_args[0][0]
        self.assertMultiLineEqual(result_log, expected_log)

        result = rules[0].literals
        self.assertEqual(expected_rule, result)

        # When
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids, strict=True)

        # Then
        with self.assertRaises(MissingInstallRequires):
            list(rules_generator.iter_rules())

    def test_missing_indirect_dependencies_package(self):
        # Given
        yaml = u"""
            packages:
              - gluon 1.0.0-1; depends (quark ^= 1.0.0)
              - atom 1.0.0-1; depends (gluon); conflicts (gdata ^= 1.0.0)
              - gdata 1.0.0-1; conflicts (atom >= 1.0.1)

            request:
              - operation: "install"
                requirement: "atom"
        """
        scenario = Scenario.from_yaml(io.StringIO(yaml))
        expected_log = ("Blocking package 'gluon 1.0.0-1' from 'remote': no"
                        " candidates found for dependency 'quark ^= 1.0.0'")
        expected_rule = (-3,)

        # When
        repos = list(scenario.remote_repositories)
        repos.append(scenario.installed_repository)
        pool = Pool(repos)
        installed_package_ids = {
            pool.package_id(p): p for p in scenario.installed_repository}
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids)
        with mock.patch('simplesat.rules_generator.logger') as mock_logger:
            rules = list(rules_generator.iter_rules())

        # Then
        result_log = mock_logger.info.call_args[0][0]
        self.assertMultiLineEqual(result_log, expected_log)

        result = [r.literals for r in rules
                  if r.reason == RuleType.package_broken][0]
        self.assertEqual(expected_rule, result)

        # When
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids, strict=True)
        with mock.patch('simplesat.rules_generator.logger') as mock_logger:
            rules = list(rules_generator.iter_rules())

        # Then
        result_log = mock_logger.warning.call_args[0][0]
        self.assertMultiLineEqual(result_log, expected_log)

        result = [r.literals for r in rules
                  if r.reason == RuleType.package_broken][0]
        self.assertEqual(expected_rule, result)

    def test_missing_direct_conflicts_package(self):
        # Given
        yaml = u"""
            packages:
              - quark 1.0.0-1
              - atom 1.0.0-1; conflicts (gdata ^= 1.0.0)

            request:
              - operation: "install"
                requirement: "atom"
        """
        scenario = Scenario.from_yaml(io.StringIO(yaml))
        expected = ("No candidates found for requirement 'gdata ^= 1.0.0',"
                    " needed for conflict with 'atom 1.0.0-1' from 'remote'")

        # When
        repos = list(scenario.remote_repositories)
        repos.append(scenario.installed_repository)
        pool = Pool(repos)
        installed_package_ids = {
            pool.package_id(p): p for p in scenario.installed_repository}
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids)
        with mock.patch('simplesat.rules_generator.logger') as mock_logger:
            list(rules_generator.iter_rules())

        # Then
        result = mock_logger.info.call_args[0][0]
        self.assertEqual(result, expected)

        # When
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids, strict=True)

        # Then
        with self.assertRaises(MissingConflicts):
            list(rules_generator.iter_rules())

    def test_allow_newer_modification(self):
        # Given
        yaml = u"""
            packages:
              - atom 1.0.1-1; depends (quark > 1.0, quark < 2.0)
              - quark 1.1.0-1
              - quark 2.1.0-1

            request:
              - operation: "install"
                requirement: "atom"
        """
        scenario = Scenario.from_yaml(io.StringIO(yaml))
        repos = list(scenario.remote_repositories)
        repos.append(scenario.installed_repository)
        pool = Pool(repos)
        installed_package_ids = {
            pool.package_id(p): p for p in scenario.installed_repository}

        # When
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids)
        rule = next(rule for rule in rules_generator.iter_rules()
                    if rule.reason == RuleType.package_requires)

        # Then
        expected = (-1, 2)
        result = rule.literals
        self.assertEqual(expected, result)

        # When
        scenario.request.allow_newer('quark')
        pool.modifiers = scenario.request.modifiers
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids)
        rule = next(rule for rule in rules_generator.iter_rules()
                    if rule.reason == RuleType.package_requires)

        # Then
        expected = (-1, 2, 3)
        result = rule.literals
        self.assertEqual(expected, result)

    def test_missing_indirect_conflicts_package(self):
        # Given
        yaml = u"""
            packages:
              - quark 1.0.0-1
              - gluon 1.0.0-1; conflicts (gdata ^= 1.0.0)
              - atom 1.0.0-1; depends (gluon);

            request:
              - operation: "install"
                requirement: "atom"
        """
        scenario = Scenario.from_yaml(io.StringIO(yaml))
        expected = ("No candidates found for requirement 'gdata ^= 1.0.0',"
                    " needed for conflict with 'gluon 1.0.0-1' from 'remote'")

        # When
        repos = list(scenario.remote_repositories)
        repos.append(scenario.installed_repository)
        pool = Pool(repos)
        installed_package_ids = {
            pool.package_id(p): p for p in scenario.installed_repository}
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids)
        with mock.patch('simplesat.rules_generator.logger') as mock_logger:
            list(rules_generator.iter_rules())

        # Then
        result = mock_logger.info.call_args[0][0]
        self.assertEqual(result, expected)

        # When
        rules_generator = RulesGenerator(
            pool, scenario.request,
            installed_package_ids=installed_package_ids, strict=True)
        with mock.patch('simplesat.rules_generator.logger') as mock_logger:
            list(rules_generator.iter_rules())

        # Then
        result = mock_logger.warning.call_args[0][0]
        self.assertEqual(result, expected)



from __future__ import absolute_import

import copy

from collections import OrderedDict


class DefaultOrderedDict(OrderedDict):
    def __init__(self, default_factory, *a, **kw):
        OrderedDict.__init__(self, *a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        return (type(self), (self.default_factory(),), None, None,
                iter(self.items()))

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        return type(self)(self.default_factory,
                          (copy.deepcopy(item) for item in self.items()))

    def __repr__(self):
        return '%s(%s, %s)' % (type(self).__name__, self.default_factory,
                               OrderedDict.__repr__(self))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from timeit import default_timer


class timed_context(object):

    def __init__(self, description):
        self.description = description
        self.elapsed = float('NaN')

    def __enter__(self):
        self.start = default_timer()
        return self

    def __exit__(self, type, value, traceback):
        self.elapsed = default_timer() - self.start

    def __str__(self):
        if self.elapsed is None:
            return "Context {} has not been run".format(self.description)
        return "ELAPSED : {} : {}".format(self.description, self.elapsed)

    def pretty(self, fmt="ELAPSED : {description} : {elapsed}"):
        return fmt.format(description=self.description, elapsed=self.elapsed)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division, print_function

from collections import OrderedDict

from attr import attr, attributes, asdict
from attr.validators import instance_of
import six

from okonomiyaki.versions import EnpkgVersion
from simplesat.constraints.kinds import (
    Any, EnpkgUpstreamMatch, Equal, Not, GEQ, GT, LEQ, LT,
)
from simplesat.constraints.requirement import InstallRequirement


MAX_BUILD = 999999999  # Nine nines... I guess


def Any_(_version):
    # This just eats the 'version' argument
    return Any()


def LEQ_LEAST_UPPER_BOUND(version):
    new_version = EnpkgVersion.from_upstream_and_build(
        str(version.upstream), MAX_BUILD)
    return LEQ(new_version)


ALLOW_NEWER_MAP = {
    Any: Any_,
    Equal: GEQ,
    Not: Not,
    GEQ: GEQ,
    GT: GT,
    LEQ: Any_,
    LT: Any_,
    EnpkgUpstreamMatch: GEQ,
}

ALLOW_OLDER_MAP = {
    Any: Any_,
    Equal: LEQ,
    Not: Not,
    GEQ: Any_,
    GT: Any_,
    LEQ: LEQ,
    LT: LT,
    EnpkgUpstreamMatch: LEQ_LEAST_UPPER_BOUND,
}

ALLOW_ANY_MAP = {
    Any: Any_,
    Equal: Any_,
    Not: Not,
    GEQ: Any_,
    GT: Any_,
    LEQ: Any_,
    LT: Any_,
    EnpkgUpstreamMatch: Any_,
}


def as_set(container):
    """ Return a set from an iterable, being careful not to disassemble
    strings.
        >>> iterable_to_set(['foo'])
        set(['foo'])
        >>> iterable_to_set('foo')
        set(['foo'])
    """
    if isinstance(container, six.string_types):
        container = (container,)
    return set(container)


_coerced_set = dict(default=(), converter=as_set,
                    validator=instance_of(set))


@attributes
class ConstraintModifiers(object):
    allow_newer = attr(**_coerced_set)
    allow_any = attr(**_coerced_set)
    allow_older = attr(**_coerced_set)

    def asdict(self):
        return {k: sorted(v) for k, v in six.iteritems(asdict(self))}

    def update(self, other_modifiers):
        """ Update modifiers with values from ConstraintModifiers instance. """
        self.allow_any.update(other_modifiers.allow_any)
        self.allow_newer.update(other_modifiers.allow_newer)
        self.allow_older.update(other_modifiers.allow_older)

    def remove(self, packages):
        """ Remove all modifiers related to `packages`.

        Parameters
        ----------
        packages : an iterable of strings
            The package names that should be completely removed.
        """
        disallowed = (type(b""), type(u""))
        if isinstance(packages, disallowed):
            raise TypeError("`packages` should be a collection, not a string.")
        packages = set(packages)
        self.allow_any.difference_update(packages)
        self.allow_newer.difference_update(packages)
        self.allow_older.difference_update(packages)

    @property
    def targets(self):
        return set.union(self.allow_newer, self.allow_any, self.allow_older)


def _modify_install_requirement(requirement, modifiers):
    """If any of the modifier rules apply, return a new Requirement with
    modified constraints, otherwise return the original requirement.
    """

    name = requirement.name
    original_constraints = constraints = requirement._constraints._constraints
    type_maps = (
        (modifiers.allow_older or (), ALLOW_OLDER_MAP),
        (modifiers.allow_newer or (), ALLOW_NEWER_MAP),
        (modifiers.allow_any or (), ALLOW_ANY_MAP),
    )

    modified = False
    for names, type_map in type_maps:
        if name in names:
            modified = True
            constraints = _modify_constraints(constraints, type_map)

    if modified and constraints != original_constraints:
        # Remove duplicate constraints
        constraints = tuple(OrderedDict.fromkeys(constraints).keys())
        return InstallRequirement(name, constraints)

    return requirement


def _modify_constraints(constraints, type_map):
    return tuple(
        type_map[type(c)](getattr(c, 'version', None))
        for c in constraints
    )


def modify_requirement(requirement, modifiers):
    if isinstance(requirement, InstallRequirement):
        return _modify_install_requirement(requirement, modifiers)
    elif requirement.has_any_version_constraint:
        new_r = _modify_install_requirement(requirement, modifiers)
        msg = "Only identity modifications are defined for {}"
        class_name = requirement.__class__.__name__
        if new_r is not requirement:
            raise NotImplementedError(msg.format(class_name))
    return requirement
